"""
SSSD In-Memory Cache (memcache) Test Cases.

:requirement: IDM-SSSD-REQ: Client side performance improvements
"""

from __future__ import annotations

import pytest
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.generic import GenericProvider
from sssd_test_framework.roles.ldap import LDAP
from sssd_test_framework.topology import KnownTopology, KnownTopologyGroup


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__lookup_users(client: Client, provider: GenericProvider):
    """
    :title: Lookup user by name uses memory cache when SSSD is stopped
    :setup:
        1. Add 'user1', 'user2' and 'user3' to SSSD
        2. Start SSSD
    :steps:
        1. Find 'user1', 'user2' and 'user3' with id(name)
        2. Check that results have correct names
        3. Stop SSSD
        4. Find 'user1', 'user2' and 'user3' with id(name)
        5. Check that results have correct names
    :expectedresults:
        1. Users are found
        2. Users have correct names
        3. SSSD is stopped
        4. Users are found
        5. Users have correct names
    :customerscenario: False
    """

    def check(users):
        for user in users:
            result = client.tools.id(user)
            assert result is not None, f"User {user} was not found using id"
            assert result.user.name == user, f"Username {result.user.name} is incorrect, {user} expected"

    users = ["user1", "user2", "user3"]
    for user in users:
        provider.user(user).add()

    client.sssd.start()

    check(users)
    client.sssd.stop()
    check(users)


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__lookup_groups(client: Client, provider: GenericProvider):
    """
    :title: Lookup group by groupname uses memory cache when SSSD is stopped
    :setup:
        1. Add 'group1', 'group2' and 'group3' to SSSD
        2. Start SSSD
    :steps:
        1. Find 'group1', 'group2' and 'group3' with getent.group(name)
        2. Check that groups have correct names
        3. Stop SSSD
        4. Find 'group1', 'group2' and 'group3' with getent.group(name)
        5. Check that groups have correct names
    :expectedresults:
        1. Groups are found
        2. Groups have correct names
        3. SSSD is stopped
        4. Groups are found
        5. Groups have correct names
    :customerscenario: False
    """

    def check(groups):
        for group in groups:
            result = client.tools.getent.group(group)
            assert result is not None, f"Group {group} was not found using getent"
            assert result.name == group, f"Groupname {result.name} is incorrect, {group} expected"

    groups = ["group1", "group2", "group3"]
    for group in groups:
        provider.group(group).add()

    client.sssd.start()

    check(groups)
    client.sssd.stop()
    check(groups)


@pytest.mark.importance("high")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__user_cache_is_disabled_and_lookup_groups(client: Client, provider: GenericProvider):
    """
    :title: Lookup group by groupname uses memory cache when SSSD is stopped and 'memcache_size_passwd' = 0
    :setup:
        1. Add 'group1', 'group2' and 'group3' to SSSD
        2. In SSSD nss change 'memcache_size_passwd' to '0'
        3. Start SSSD
    :steps:
        1. Find 'group1', 'group2' and 'group3' with getent.group(name)
        2. Check that groups have correct names
        3. Stop SSSD
        4. Find 'group1', 'group2' and 'group3' with getent.group(name)
        5. Check that groups have correct names
    :expectedresults:
        1. Groups are found
        2. Groups have correct names
        3. SSSD is stopped
        4. Groups are found
        5. Groups have correct names
    :customerscenario: False
    """

    def check(groups):
        for group in groups:
            result = client.tools.getent.group(group)
            assert result is not None, f"Group {group} was not found using getent"
            assert result.name == group, f"Groupname {result.name} is incorrect, {group} expected"

    groups = ["group1", "group2", "group3"]
    for group in groups:
        provider.group(group).add()

    client.sssd.nss["memcache_size_passwd"] = "0"
    client.sssd.start()

    check(groups)
    client.sssd.stop()
    check(groups)


@pytest.mark.importance("high")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__user_cache_is_disabled_and_lookup_users(client: Client, provider: GenericProvider):
    """
    :title: Lookup user by name when SSSD is stopped and 'memcache_size_passwd' = 0
            uses memory cache therefore user is not found
    :setup:
        1. Add users to SSSD
        2. Set users uids and gids
        3. In SSSD nss change 'memcache_size_passwd' to '0'
        4. Start SSSD
    :steps:
        1. Find 'user1', 'user2' and 'user3' with id(name)
        2. Check that users have correct names
        3. Stop SSSD
        4. Find users with id(name)
        5. Find users with id(uid)
    :expectedresults:
        1. Users are found
        2. Users have correct names
        3. SSSD is stopped
        4. Users are not found
        5. Users are not found
    :customerscenario: False
    """
    ids = [("user1", 10001), ("user2", 10002), ("user3", 10003)]
    for user, id in ids:
        provider.user(user).add(uid=id, gid=id + 500)

    client.sssd.nss["memcache_size_passwd"] = "0"
    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    for user, id in ids:
        result = client.tools.id(user)
        assert result is not None, f"User {user} was not found using id"
        assert result.user.name == user, f"Username {result.user.name} is incorrect, {user} expected"
        assert result.user.id == id, f"User id {result.user.id} is incorrect, {id} expected"

    client.sssd.stop()

    for user, id in ids:
        assert client.tools.id(user) is None, f"User {user} was found which is not expected"
        assert client.tools.id(id) is None, f"User with id {id} was found which is not expected"


@pytest.mark.importance("high")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__initgroup_cache_is_disabled_and_lookup_groups(client: Client, provider: GenericProvider):
    """
    :title: Lookup group by groupname when SSSD is stopped and 'memcache_size_initgroups' = 0 uses memory cache
    :setup:
        1. Add 'group1', 'group2' and 'group3' to SSSD
        2. In SSSD nss change 'memcache_size_initgroups' to '0'
        3. Start SSSD
    :steps:
        1. Find 'group1', 'group2' and 'group3' with getent.group(name)
        2. Check that groups have correct names
        3. Stop SSSD
        4. Find 'group1', 'group2' and 'group3' with getent.group(name)
        5. Check that groups have correct names
    :expectedresults:
        1. Groups are found
        2. Groups have correct names
        3. SSSD is stopped
        4. Groups are found
        5. Groups have correct names
    :customerscenario: False
    """

    def check(groups):
        for group in groups:
            result = client.tools.getent.group(group)
            assert result is not None, f"Group {group} was not found using getent"
            assert result.name == group, f"Groupname {result.name} is incorrect, {group} expected"

    groups = ["group1", "group2", "group3"]
    for group in groups:
        provider.group(group).add()

    client.sssd.nss["memcache_size_initgroups"] = "0"
    client.sssd.start()

    check(groups)
    client.sssd.stop()
    check(groups)


@pytest.mark.importance("high")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__initgroup_cache_is_disabled_and_lookup_users(client: Client, provider: GenericProvider):
    """
    :title: Lookup user by name and id when SSSD is stopped and 'memcache_size_initgroups' = 0 uses memory cache
    :setup:
        1. Add users to SSSD
        2. Set users uids and ids
        3. In SSSD nss change 'memcache_size_initgroups' to '0'
        4. Start SSSD
    :steps:
        1. Find 'user1', 'user2' and 'user3' with id(name)
        2. Check that users have correct names and uids
        3. Stop SSSD
        4. Find 'user1', 'user2' and 'user3' with id(name)
        5. Check that users have correct names and uids
        6. Find 'user1', 'user2' and 'user3' with id(uid)
        7. Check that users have correct names and uids
    :expectedresults:
        1. Users are found
        2. Users have correct names and uids
        3. SSSD is stopped
        4. Users are found
        5. Users have correct names and uids
        6. Users are found
        7. Users have correct names and uids
    :customerscenario: False
    """

    def check(ids):
        for name, id in ids:
            result = client.tools.id(name)
            assert result is not None, f"User {name} was not found using id"
            assert result.user.name == name, f"Username {result.user.name} is incorrect, {user} expected"
            assert result.user.id == id, f"User id {result.user.id} is incorrect, {id} expected"

            result = client.tools.id(id)
            assert result is not None, f"User with id {id} was not found using id"
            assert result.user.name == name, f"Username {result.user.name} is incorrect, {user} expected"
            assert result.user.id == id, f"User id {result.user.id} is incorrect, {id} expected"

    ids = [("user1", 10001), ("user2", 10002), ("user3", 10003)]
    for user, id in ids:
        provider.user(user).add(uid=id, gid=id + 500)

    client.sssd.nss["memcache_size_initgroups"] = "0"
    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    check(ids)
    client.sssd.stop()
    check(ids)


@pytest.mark.importance("high")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__group_cache_disabled_and_lookup_groups(client: Client, provider: GenericProvider):
    """
    :title: Lookup user by name and id when SSSD is stopped and 'memcache_size_group' = 0 uses memory cache,
            but lookup groups is not possible
    :setup:
        1. Add users to SSSD
        2. Set users uids and gids
        3. Add groups to SSSD
        4. Set groups gids
        5. Add users to groups
        6. In SSSD nss change 'memcache_size_group' to '0'
        7. Start SSSD
    :steps:
        1. Find 'user1', 'user2' and 'user3' with id(name)
        2. Check that users have correct names
        3. Find 'group1' and 'group2' by getent.group(gid)
        4. Check that groups have correct gids and members
        5. Stop SSSD
        6. Find 'user1', 'user2' and 'user3' with id(name)
        7. Check that users have correct names
        8. Find 'group1' and 'group2' by getent.group(name)
        9. Find 'group1' and 'group2' by getent.group(gid)
    :expectedresults:
        1. Users are found
        2. Users have correct names
        3. Groups are found
        4. Groups have correct gids and members
        5. SSSD is stopped
        6. Users are found
        7. Users have correct names
        8. Groups are not found
        9. Groups are not found
    :customerscenario: False
    """

    def check(users):
        for user in users:
            rUser = client.tools.id(user)
            assert rUser is not None, f"User {rUser} was not found using id"
            assert rUser.user.name == user, f"Username {rUser.user.name} is incorrect, {user} expected"

    u1 = provider.user("user1").add(uid=10001, gid=19001)
    u2 = provider.user("user2").add(uid=10002, gid=19002)
    u3 = provider.user("user3").add(uid=10003, gid=19003)

    provider.group("group1").add(gid=1111).add_member(u1)
    provider.group("group2").add(gid=2222).add_members([u1, u2, u3])

    client.sssd.nss["memcache_size_group"] = "0"
    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    users = ["user1", "user2", "user3"]
    check(users)

    for group, members in [(1111, ["user1"]), (2222, ["user1", "user2", "user3"])]:
        result = client.tools.getent.group(group)
        assert result is not None, f"Group {group} was not found using getent"
        assert result.gid == group, f"Group gid {result.gid} is incorrect, {group} expected"
        assert result.members == members, f"Group {group} members did not match the expected ones"

    client.sssd.stop()

    check(users)

    assert client.tools.id("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.id("group2") is None, "Group group2 was found which is not expected"
    assert client.tools.id(1111) is None, "Group with gid 1111 was found which is not expected"
    assert client.tools.id(2222) is None, "Group with gid 2222 was found which is not expected"


@pytest.mark.importance("high")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__all_caches_disabled_and_all_lookups_fails(client: Client, provider: GenericProvider):
    """
    :title: Lookup user and group when SSSD is stopped and whole cache disabled
            uses memory cache and therefore it is not possible
    :setup:
        1. Add users to SSSD
        2. Set users uids
        3. Add groups to SSSD
        4. Set groups gids
        5. Add users to groups
        6. In SSSD nss change 'memcache_size_passwd' to '0'
        7. In SSSD nss change 'memcache_size_group' to '0'
        8. In SSSD nss change 'memcache_size_initgroups' to '0'
        9. Start SSSD
    :steps:
        1. Find 'user1', 'user2' and 'user3' with id(name)
        2. Check that users have correct names
        3. Find 'group1' and 'group2' by getent.group(name)
        4. Check that groups have correct names and members
        5. Stop SSSD
        6. Find 'user1', 'user2' and 'user3' with id(name)
        7. Find 'user1', 'user2' and 'user3' with id(uid)
        8. Find 'group1' and 'group2' by getent.group(name)
        9. Find 'group1' and 'group2' by getent.group(gid)
    :expectedresults:
        1. Users are found
        2. Users have correct names
        3. Groups are found
        4. Groups have correct names and members
        5. SSSD is stopped
        6. Users are not found
        7. Users are not found
        8. Groups are not found
        9. Groups are not found
    :customerscenario: False
    """
    u1 = provider.user("user1").add(uid=10001, gid=19001)
    u2 = provider.user("user2").add(uid=10002, gid=19002)
    u3 = provider.user("user3").add(uid=10003, gid=19003)

    provider.group("group1").add(gid=1111).add_member(u1)
    provider.group("group2").add(gid=2222).add_members([u1, u2, u3])

    client.sssd.nss["memcache_size_passwd"] = "0"
    client.sssd.nss["memcache_size_group"] = "0"
    client.sssd.nss["memcache_size_initgroups"] = "0"
    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    for user in ["user1", "user2", "user3"]:
        result = client.tools.id(user)
        assert result is not None, f"User {user} was not found using id"
        assert result.user.name == user, f"Username {result.user.name} is incorrect, {user} expected"

    for group, members in [("group1", ["user1"]), ("group2", ["user1", "user2", "user3"])]:
        gresult = client.tools.getent.group(group)
        assert gresult is not None, f"Group {group} was not found using id"
        assert gresult.name == group, f"Groupname {gresult.name} is incorrect, {group} expected"
        assert gresult.members == members, f"Group {group} members did not match the expected ones"

    client.sssd.stop()

    for user in ["user1", "user2", "user3"]:
        assert client.tools.id(user) is None, f"User {user} was found which is not expected"

    for id in [10001, 10002, 10003]:
        assert client.tools.id(id) is None, f"User with id {id} was found which is not expected"

    assert client.tools.getent.group("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.getent.group("group2") is None, "Group group2 was found which is not expected"
    assert client.tools.getent.group(1111) is None, "Group with gid 1111 was found which is not expected"
    assert client.tools.getent.group(2222) is None, "Group with gid 2222 was found which is not expected"


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__lookup_users_check_group_memberships(client: Client, provider: GenericProvider):
    """
    :title: Lookup user by name and test membership by name use memory cache when SSSD is stopped
    :setup:
        1. Add 'user1', 'user2' and 'user3' to SSSD
        2. Add 'group1' and 'group2' to SSSD
        3. Add users to groups
        4. Start SSSD
    :steps:
        1. Find 'user1', 'user2' and 'user3' with id(name)
        2. Check that users are members of correct groups
        3. Stop SSSD
        4. Find 'user1', 'user2' and 'user3' with id(name)
        5. Check that users are members of correct groups
    :expectedresults:
        1. Users are found
        2. Users are members of correct groups
        3. SSSD is stopped
        4. Users are found
        5. Users are members of correct groups
    :customerscenario: False
    """

    def check():
        result = client.tools.id("user1")
        assert result is not None, "User user1 was not found using id"
        assert result.memberof(["group1", "group2"]), "User user1 is member of incorrect groups"

        result = client.tools.id("user2")
        assert result is not None, "User user2 was not found using id"
        assert result.memberof(["group2"]), "User user2 is member of incorrect groups"

        result = client.tools.id("user3")
        assert result is not None, "User user3 was not found using id"
        assert result.memberof(["group2"]), "User user3 is member of incorrect groups"

    u1 = provider.user("user1").add()
    u2 = provider.user("user2").add()
    u3 = provider.user("user3").add()

    provider.group("group1").add().add_member(u1)
    provider.group("group2").add().add_members([u1, u2, u3])

    client.sssd.start()

    check()
    client.sssd.stop()
    check()


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__lookup_users_and_check_membership_by_gid(client: Client, provider: GenericProvider):
    """
    :title: Lookup user by name and test membership by gid use memory cache when SSSD is stopped
    :setup:
        1. Add 'user1', 'user2' and 'user3' to SSSD
        2. Add 'group1', 'group2' and 'group3' to SSSD
        3. Set group gids
        4. Add users to groups
        5. Start SSSD
    :steps:
        1. Find 'user1', 'user2' and 'user3' with id(name)
        2. Check that users are members of correct groups
        3. Stop SSSD
        4. Find 'user1', 'user2' and 'user3' with id(name)
        5. Check that users are members of correct groups
    :expectedresults:
        1. Users are found
        2. Users are members of correct groups
        3. SSSD is stopped
        4. Users are found
        5. Users are members of correct groups
    :customerscenario: False
    """

    def check():
        result = client.tools.id("user1")
        assert result is not None, "User user1 was not found using id"
        assert result.memberof([1001, 1002]), "User user1 is member of incorrect groups"

        result = client.tools.id("user2")
        assert result is not None, "User user2 was not found using id"
        assert result.memberof([1002]), "User user2 is member of incorrect groups"

        result = client.tools.id("user3")
        assert result is not None, "User user3 was not found using id"
        assert result.memberof([1002]), "User user3 is member of incorrect groups"

    u1 = provider.user("user1").add(uid=11001, gid=19001)
    u2 = provider.user("user2").add(uid=11002, gid=19002)
    u3 = provider.user("user3").add(uid=11003, gid=19003)

    provider.group("group1").add(gid=1001).add_member(u1)
    provider.group("group2").add(gid=1002).add_members([u1, u2, u3])
    provider.group("group3").add(gid=1003)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    check()
    client.sssd.stop()
    check()


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__lookup_uids_and_check_membership_by_gid(client: Client, provider: GenericProvider):
    """
    :title: Lookup user by id and test membership by gid use memory cache when SSSD is stopped
    :setup:
        1. Add 'user1', 'user2' and 'user3' to SSSD
        2. Set users uids and gids
        3. Add 'group1' and 'group2' to SSSD
        4. Set groups gids
        5. Add users to groups
        6. Start SSSD
    :steps:
        1. Find users by id(uid)
        2. Check that users are members of correct groups
        3. Stop SSSD
        4. Find users by id(uid)
        5. Check that users are members of correct groups
    :expectedresults:
        1. Users are found
        2. Users are members of correct groups
        3. SSSD is stopped
        4. Users are found
        5. Users are members of correct groups
    :customerscenario: False
    """

    def check():
        result = client.tools.id(2001)
        assert result is not None, "User with id 2001 was not found using id"
        assert result.memberof([101, 1001, 1002]), "User with id 2001 is member of incorrect groups"

        result = client.tools.id(2002)
        assert result is not None, "User with id 2002 was not found using id"
        assert result.memberof([102, 1002]), "User with id 2002 is member of incorrect groups"

        result = client.tools.id(2003)
        assert result is not None, "User with id 2003 was not found using id"
        assert result.memberof([103, 1002]), "User with id 2003 is member of incorrect groups"

    u1 = provider.user("user1").add(uid=2001, gid=101)
    u2 = provider.user("user2").add(uid=2002, gid=102)
    u3 = provider.user("user3").add(uid=2003, gid=103)

    provider.group("group1").add(gid=1001).add_member(u1)
    provider.group("group2").add(gid=1002).add_members([u1, u2, u3])

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    check()
    client.sssd.stop()
    check()


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__lookup_users_by_fully_qualified_names(client: Client, provider: GenericProvider):
    """
    :title: Lookup user by full name when 'use_fully_qualified_names' is 'true'
            uses memory cache when sssd is stopped
    :setup:
        1. Add 'user1' and 'user2' to SSSD
        2. In SSSD domain change 'use_fully_qualified_names' to 'true'
        3. Start SSSD
    :steps:
        1. Find 'user1' and 'user2' with id(name)
        2. Find 'user1' and 'user2' with id(name@domain)
        3. Check that users have correct full names
        4. Stop SSSD
        5. Find 'user1' and 'user2' with id(name)
        6. Find 'user1' and 'user2' with id(name@domain)
        7. Check that users have correct full names
    :expectedresults:
        1. Users are not found
        2. Users are found
        3. Users have correct full names
        4. SSSD is stopped
        5. Users are not found
        6. Users are found
        7. Users have correct full names
    :customerscenario: False
    """

    def check():
        assert client.tools.id("user1") is None, "User user1 should not be found without fully qualified name"
        assert client.tools.id("user2") is None, "User user2 should not be found without fully qualified name"

        result = client.tools.id("user1@test")
        assert result is not None, "User user1@test was not found using id"
        assert result.user.name == "user1@test", f"User {result.user.name} has incorrect name, user1@test expected"

        result = client.tools.id("user2@test")
        assert result is not None, "User user2@test was not found using id"
        assert result.user.name == "user2@test", f"User {result.user.name} has incorrect name, user2@test expected"

    provider.user("user1").add()
    provider.user("user2").add()

    client.sssd.domain["use_fully_qualified_names"] = "true"
    client.sssd.start()

    check()
    client.sssd.stop()
    check()


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__lookup_users_when_case_insensitive_is_false(client: Client, provider: GenericProvider):
    """
    :title: Lookup user by case insensitive name when 'case_sensitive' is 'false'
            uses memory cache when SSSD is stopped
    :setup:
        1. Add 'user1' to SSSD
        2. Set user gid and uid
        3. Add groups to SSSD
        4. Set group gid
        5. Add member to the groups
        6. In SSSD domain change 'case_sensitive' to 'false'
        7. Start SSSD
    :steps:
        1. Find users with getent.initgroups(name), where name is in random lower and upper case format
        2. Check that usernames are correct
        3. Check that user is member of correct groups
        4. Stop SSSD
        5. Find user with getent.initgroups(name), where name is last name used when resolving user
        6. Check that username is correct
        7. Check that user is member of correct groups
        8. Find users with getent.initgroups(name), where names are previously used names
    :expectedresults:
        1. Users are found
        2. Users have correct names
        3. User is member of correct groups
        4. SSSD is stopped
        5. User is found
        6. User has correct name
        7. User is member of correct groups
        8. Users are not found
    :customerscenario: False
    """
    u1 = provider.user("user1").add(uid=10001, gid=2001)
    provider.group("group1").add(gid=10010).add_member(u1)
    provider.group("group2").add(gid=10011).add_member(u1)
    provider.group("group3").add(gid=10012).add_member(u1)

    client.sssd.domain["case_sensitive"] = "false"
    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    for name in ["uSer1", "useR1", "USER1", "user1", "uSER1"]:
        result = client.tools.getent.initgroups(name)
        assert result.name == name, f"Username {result.name} is not correct, {name} expected"
        assert result.memberof([10010, 10011, 10012]), f"User {result.name} is member of wrong groups"

    client.sssd.stop()

    result = client.tools.getent.initgroups("uSER1")
    assert result.name == "uSER1", f"Username {result.name} is not correct, uSER1 expected"
    assert result.memberof([10010, 10011, 10012]), f"User {result.name} is member of wrong groups"

    # Only last version of name is stored in cache
    # That is why initgroups call returns no secondary groups
    assert client.tools.getent.initgroups("uSer1").groups == [], "User uSer1 should not be found in cache"
    assert client.tools.getent.initgroups("useR1").groups == [], "User useR1 should not be found in cache"
    assert client.tools.getent.initgroups("USER1").groups == [], "User USER1 should not be found in cache"
    assert client.tools.getent.initgroups("user1").groups == [], "User user1 should not be found in cache"


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__lookup_users_when_fully_qualified_name_is_true_and_case_ins_is_false(
    client: Client, provider: GenericProvider
):
    """
    :title: Lookup user by case insensitive fully qualified name when 'case_sensitive' is 'false'
            and 'use_fully_qualified_names' is 'true' uses memory cache when SSSD is stopped
    :setup:
        1. Add user to SSSD
        2. Add groups to SSSD
        3. Set groups gids
        4. Add members to the groups
        5. In SSSD domain change 'use_fully_qualified_names' to 'true'
        6. In SSSD domain change 'case_sensitive' to 'false'
        7. Start SSSD
    :steps:
        1. Find user with names without domain
        2. Find user with getent.initgroups(name@domain), where name is in random lower and upper case format
        3. Check that user is members of correct groups
        4. Stop SSSD
        5. Find user with getent.initgroups(name@domain), where same name as in 2.
        6. Check that user is member of correct groups
        7. Find users with names, that should not be found
    :expectedresults:
        1. User is not found
        2. User is found
        3. User is member of correct groups
        4. SSSD is stopped
        5. User is found
        6. User is member of correct groups
        7. Users are not found
    :customerscenario: False
    """
    u1 = provider.user("user1").add(gid=19001, uid=11001)

    provider.group("group1").add(gid=20001).add_member(u1)
    provider.group("group2").add(gid=20002).add_member(u1)
    provider.group("group3").add(gid=20003).add_member(u1)

    client.sssd.domain["use_fully_qualified_names"] = "true"
    client.sssd.domain["case_sensitive"] = "false"
    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    assert client.tools.getent.initgroups("uSer1").groups == [], "User uSer1 should be found only with fq name"
    assert client.tools.getent.initgroups("user1").groups == [], "User user1 should be found only with fq name"

    result = client.tools.getent.initgroups("uSer1@test")
    assert result.memberof([20001, 20002, 20003]), "User uSer1@test is member of incorrect groups"

    client.sssd.stop()

    result = client.tools.getent.initgroups("uSer1@test")
    assert result.memberof([20001, 20002, 20003]), "User uSer1@test is member of incorrect groups"

    assert client.tools.getent.initgroups("user1@test").groups == [], "User user1@test should not be found in cache"
    assert client.tools.getent.initgroups("user1").groups == [], "User user1 should be found only with fq name"
    assert client.tools.getent.initgroups("uSer1").groups == [], "User uSer1 should be found only with fq name"


@pytest.mark.importance("high")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__invalidation_of_gids_after_initgroups(client: Client, provider: GenericProvider):
    """
    :title: Invalidate groups after initgroups call when SSSD is stopped
    :setup:
        1. Add 'user1' to SSSD
        2. Set user uid and gid
        3. Add groups to SSSD
        4. Set groups gids
        5. Add members to the groups
        6. Start SSSD
    :steps:
        1. Check that "user1" has correct attributes
        2. Check that groups have correct attributes
        3. Check that "user1" has correct initgroups
        4. Stop SSSD
        5. Check that "group2" has correct attributes
        6. Check that "user1" has correct initgroups
        7. All "user1" initgroups should be invalidated and not found
    :expectedresults:
        1. User has correct attributes
        2. Groups have correct attributes
        3. User has correct initgroups
        4. SSSD is stopped
        5. Group has correct attributes
        6. User has correct attributes
        7. Groups are not found
    :customerscenario: False
    """

    def check_user_passwd():
        for user in ("user1", 10001):
            result = client.tools.getent.passwd(user)
            assert result is not None, f"User {user} was not found using getent"
            assert result.uid == 10001, f"User id {result.uid} is incorrect, expected 10001"
            assert result.name == "user1", f"Username {result.name} is incorrect, expected user1"

    def check_initgroups():
        result = client.tools.getent.initgroups("user1")
        assert result.name == "user1", f"Username {result.name} is incorrect, user1 expected"
        assert result.memberof([12345]), "User user1 is member of incorrect groups"

    def check_group(name, gid):
        for group in (name, gid):
            gresult = client.tools.getent.group(group)
            assert gresult is not None, f"Group {group} was not found using getent"
            assert gresult.name == name, f"Groupname {gresult.name} is incorrect, {name} expected"
            assert gresult.gid == gid, f"Group gid {gresult.gid} is incorrect, {gid} expected"

    u1 = provider.user("user1").add(uid=10001, gid=19001)

    provider.group("group1").add(gid=12345).add_member(u1)
    provider.group("group1_").add(gid=123450).add_member(u1)
    provider.group("group2").add(gid=22222)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    check_user_passwd()

    check_group("group1", 12345)
    check_group("group2", 22222)

    check_initgroups()

    client.sssd.stop()

    check_user_passwd()
    check_group("group2", 22222)

    check_initgroups()

    assert client.tools.getent.group(12345) is None, "Group with gid 12345 was found which is not expected"
    assert client.tools.getent.group(123450) is None, "Group with gid 123450 was found which is not expected"
    assert client.tools.getent.group("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.getent.group("group1_") is None, "Group group1_ was found which is not expected"


@pytest.mark.importance("high")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__lookup_initgroups_without_change_in_membership(client: Client, provider: GenericProvider):
    """
    :title: Invalidated cache, after refresh and stopped SSSD, has everything loaded in memory
    :setup:
        1. Add 'user1' to SSSD
        2. Set user gid and uid
        3. Add 'group1' to SSSD
        4. Set group gid
        5. Add members to the group
        6. Start SSSD
    :steps:
        1. Find user with id(name) and id(uid)
        2. Check that user is member of correct groups
        3. Find group with getent.group(name) and getent.group(gid)
        4. Check that the group have correct name and gid
        5. Invalidate whole cache
        6. Find user with id(name) and id(uid)
        7. Check that user is member of correct groups
        8. Find group with getent.group(name) and getent.group(gid)
        9. Check that the group have correct name and gid
        10. Stop SSSD
        11. Find user with id(name) and id(uid)
        12. Check that user is member of correct groups
        13. Find group with getent.group(name) and getent.group(gid)
        14. Check that the group have correct name and gid
    :expectedresults:
        1. User is found
        2. User is member of correct groups
        3. Group is found
        4. Group has correct name and gid
        5. Cache is invalidated
        6. User is found
        7. User is member of correct groups
        8. Group is found
        9. Group has correct name and gid
        10. SSSD is stopped
        11. User is found
        12. User is member of correct groups
        13. Group is found
        14. Group has correct name and gid
    :customerscenario: False
    """

    def check():
        result = client.tools.id("user1")
        assert result is not None, "User user1 was not found using id"
        assert result.memberof([111, 12345]), "User user1 is member of incorrect groups"

        result = client.tools.id(10001)
        assert result is not None, "User with id 10001 was not found using id"
        assert result.memberof([111, 12345]), "User with id 10001 is member of incorrect groups"

        gresult = client.tools.getent.group("group1")
        assert gresult is not None, "Group group1 was not found using getent"
        assert gresult.gid == 12345, f"Group gid {gresult.gid} is incorrect, 12345 expected"

        gresult = client.tools.getent.group(12345)
        assert gresult is not None, "Group with gid 12345 was not found using getent"
        assert gresult.name == "group1", f"Groupname {gresult.name} is incorrect, group1 expected"

        gresult = client.tools.getent.group("group2")
        assert gresult is not None, "Group group2 was not found using getent"
        assert gresult.gid == 222222, f"Group gid {gresult.gid} is incorrect, 222222 expected"

        gresult = client.tools.getent.group(222222)
        assert gresult is not None, "Group with gid 222222 was not found using getent"
        assert gresult.name == "group2", f"Groupname {gresult.name} is incorrect, group2 expected"

        result = client.tools.getent.initgroups("user1")
        assert result.memberof([12345, 123450]), "User user1 is member of incorrect groups"

    u1 = provider.user("user1").add(uid=10001, gid=111)
    provider.group("group1").add(gid=12345).add_member(u1)
    provider.group("group1_").add(gid=123450).add_member(u1)
    provider.group("group2").add(gid=222222)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    check()
    client.sssctl.cache_expire(everything=True)
    check()
    client.sssd.stop()
    check()


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__invalidate_user_cache_before_stop(client: Client, provider: GenericProvider):
    """
    :title: Invalidate user cache before SSSD is stopped
    :setup:
        1. Add 'user1' and 'user2' to SSSD
        2. Set users gids and uids
        3. Add 'group1' and 'group2' to SSSD
        4. Set groups gids
        5. Add members to the groups
        6. Start SSSD
    :steps:
        1. Find user with id(name)
        2. Check that user has correct id
        3. Check that user is member of correct groups
        4. Invalidate cache for 'user1'
        5. Stop SSSD
        6. Find user by id(name) and id(uid)
        7. Find the user's groups by getent.group(name) and getent.group(uid)
    :expectedresults:
        1. User is found
        2. User has correct id
        3. User is member of correct groups
        4. Cache is invalidated
        5. SSSD is stopped
        6. User is not found
        7. Group is not found
    :customerscenario: False
    """
    u1 = provider.user("user1").add(uid=123456, gid=110011)
    u2 = provider.user("user2").add(uid=220022, gid=222222)

    provider.group("group1").add(gid=101010).add_member(u1)
    provider.group("group2").add(gid=202020).add_members([u1, u2])

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    result = client.tools.id("user1")
    assert result is not None, "User user1 was not found using id"
    assert result.user.id == 123456, f"User id {result.user.id} is incorrect, 123456 expected"
    assert result.memberof([110011, 101010, 202020]), "User user1 is member of incorrect groups"

    client.sssctl.cache_expire(user="user1")
    client.sssd.stop()

    assert client.tools.id("user1") is None, "User user1 was found which is not expected"
    assert client.tools.id(123456) is None, "User with id 123456 was found which is not expected"
    assert client.tools.getent.group("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.getent.group(110011) is None, "Group with gid 110011 was found which is not expected"
    assert client.tools.getent.group("group2") is None, "Group group2 was found which is not expected"
    assert client.tools.getent.group(202020) is None, "Group with gid 202020 was found which is not expected"


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__invalidate_user_cache_after_stop(client: Client, provider: GenericProvider):
    """
    :title: Invalidate user cache after SSSD is stopped
    :setup:
        1. Add 'user1' and 'user2' to SSSD
        2. Set users gids and uids
        3. Add 'group1' and 'group2' to SSSD
        4. Set groups gids
        5. Add members to the groups
        6. Start SSSD
    :steps:
        1. Find user with id(name)
        2. Check that user has correct id
        3. Check that user is member of correct groups
        4. Stop SSSD
        5. Invalidate cache for 'user1'
        6. Find user by id(name) and id(uid)
        7. Find the user's groups by getent.group(name) and getent.group(uid)
    :expectedresults:
        1. User is found
        2. User has correct id
        3. User is member of correct groups
        4. SSSD is stopped
        5. Cache is invalidated
        6. User is not found
        7. Group is not found
    :customerscenario: False
    """
    u1 = provider.user("user1").add(uid=123456, gid=110011)
    u2 = provider.user("user2").add(uid=220022, gid=222222)

    provider.group("group1").add(gid=101010).add_member(u1)
    provider.group("group2").add(gid=202020).add_members([u1, u2])

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    result = client.tools.id("user1")
    assert result is not None, "User user1 was not found using id"
    assert result.user.id == 123456, f"User id {result.user.id} is incorrect, 123456 expected"
    assert result.memberof([110011, 101010, 202020]), "User user1 is member of incorrect groups"

    client.sssd.stop()
    client.sssctl.cache_expire(user="user1")

    assert client.tools.id("user1") is None, "User user1 was found which is not expected"
    assert client.tools.id(123456) is None, "User with id 123456 was found which is not expected"
    assert client.tools.getent.group("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.getent.group(110011) is None, "Group with gid 110011 was found which is not expected"
    assert client.tools.getent.group("group2") is None, "Group group2 was found which is not expected"
    assert client.tools.getent.group(202020) is None, "Group with gid 202020 was found which is not expected"


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__invalidate_users_cache_before_stop(client: Client, provider: GenericProvider):
    """
    :title: Invalidate users cache before SSSD is stopped
    :setup:
        1. Add 'user1' and 'user2' to SSSD
        2. Set users gids and uids
        3. Add 'group1' and 'group2' to SSSD
        4. Set groups gids
        5. Add members to the groups
        6. Start SSSD
    :steps:
        1. Find users with id(name)
        2. Check that users have correct ids
        3. Check that users are members of correct groups
        4. Invalidate cache for all users
        5. Stop SSSD
        6. Find users by id(name) and id(uid)
        7. Find the groups of users by getent.group(name) and getent.group(uid)
    :expectedresults:
        1. Users are found
        2. Users have correct ids
        3. Users are members of correct groups
        4. Cache is invalidated
        5. SSSD is stopped
        6. Users are not found
        7. Groups are not found
    :customerscenario: False
    """
    u1 = provider.user("user1").add(uid=123456, gid=110011)
    u2 = provider.user("user2").add(uid=220022, gid=222222)

    provider.group("group1").add(gid=101010).add_member(u1)
    provider.group("group2").add(gid=202020).add_members([u1, u2])

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    result = client.tools.id("user1")
    assert result is not None, "User user1 was not found using id"
    assert result.user.id == 123456, f"User id {result.user.id} is incorrect, 123456 expected"
    assert result.memberof([110011, 101010, 202020]), "User user1 is member of incorrect groups"

    result = client.tools.id("user2")
    assert result is not None, "User user2 was not found using id"
    assert result.user.id == 220022, f"User id {result.user.id} is incorrect, 220022 expected"
    assert result.memberof([222222, 202020]), "User user2 is member of incorrect groups"

    client.sssctl.cache_expire(users=True)
    client.sssd.stop()

    assert client.tools.id("user1") is None, "User user1 was found which is not expected"
    assert client.tools.id(123456) is None, "User with id 123456 was found which is not expected"
    assert client.tools.getent.group("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.getent.group(110011) is None, "Group with gid 110011 was found which is not expected"
    assert client.tools.getent.group("group2") is None, "Group group2 was found which is not expected"
    assert client.tools.getent.group(202020) is None, "Group with gid 202020 was found which is not expected"
    assert client.tools.id("user2") is None, "User user2 was found which is not expected"
    assert client.tools.id(220022) is None, "User with id 220022 was found which is not expected"
    assert client.tools.getent.group(222222) is None, "Group with gid 222222 was found which is not expected"


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__invalidate_users_cache_after_stop(client: Client, provider: GenericProvider):
    """
    :title: Invalidate users cache after SSSD is stopped
    :setup:
        1. Add 'user1' and 'user2' to SSSD
        2. Set users gids and uids
        3. Add 'group1' and 'group2' to SSSD
        4. Set groups gids
        5. Add members to the groups
        6. Start SSSD
    :steps:
        1. Find users with id(name)
        2. Check that users have correct ids
        3. Check that users are members of correct groups
        4. Stop SSSD
        5. Invalidate cache for all users
        6. Find users by id(name) and id(uid)
        7. Find the groups of users by getent.group(name) and getent.group(uid)
    :expectedresults:
        1. Users are found
        2. Users have correct ids
        3. Users are members of correct groups
        4. SSSD is stopped
        5. Cache is invalidated
        6. Users are not found
        7. Groups are not found
    :customerscenario: False
    """
    u1 = provider.user("user1").add(uid=123456, gid=110011)
    u2 = provider.user("user2").add(uid=220022, gid=222222)

    provider.group("group1").add(gid=101010).add_member(u1)
    provider.group("group2").add(gid=202020).add_members([u1, u2])

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    result = client.tools.id("user1")
    assert result is not None, "User user1 was not found using id"
    assert result.user.id == 123456, f"User id {result.user.id} is incorrect, 123456 expected"
    assert result.memberof([110011, 101010, 202020]), "User user1 is member of incorrect groups"

    result = client.tools.id("user2")
    assert result is not None, "User user2 was not found using id"
    assert result.user.id == 220022, f"User id {result.user.id} is incorrect, 220022 expected"
    assert result.memberof([222222, 202020]), "User user2 is member of incorrect groups"

    client.sssd.stop()
    client.sssctl.cache_expire(users=True)

    assert client.tools.id("user1") is None, "User user1 was found which is not expected"
    assert client.tools.id(123456) is None, "User with id 123456 was found which is not expected"
    assert client.tools.getent.group("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.getent.group(110011) is None, "Group with gid 110011 was found which is not expected"
    assert client.tools.getent.group("group2") is None, "Group group2 was found which is not expected"
    assert client.tools.getent.group(202020) is None, "Group with gid 202020 was found which is not expected"
    assert client.tools.id("user2") is None, "User user2 was found which is not expected"
    assert client.tools.id(220022) is None, "User with id 220022 was found which is not expected"
    assert client.tools.getent.group(222222) is None, "Group with gid 222222 was found which is not expected"


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__invalidate_group_cache_before_stop(client: Client, provider: GenericProvider):
    """
    :title: Invalidate group cache before SSSD is stopped
    :setup:
        1. Add 'group1' to SSSD
        2. Set group gid
        3. Start SSSD
    :steps:
        1. Find the 'group1' getent.group(name)
        2. Check that group has correct id
        3. Check that group has correct name
        4. Invalidate cache for 'group1'
        5. Stop SSSD
        6. Find the 'group1' getent.group(name) and getent.group(uid)
    :expectedresults:
        1. Group is found
        2. Group has correct id
        3. Group has correct name
        4. Cache is invalidated
        5. SSSD is stopped
        6. Group is not found
    :customerscenario: False
    """
    provider.group("group1").add(gid=101010)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    result = client.tools.getent.group("group1")
    assert result is not None, "Group group1 was not found using getent"
    assert result.name == "group1", f"Groupname {result.name} is incorrect, group1 expected"
    assert result.gid == 101010, f"Group gid {result.gid} is incorrect, 101010 expected"

    client.sssctl.cache_expire(group="group1")
    client.sssd.stop()

    assert client.tools.getent.group("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.getent.group(110011) is None, "Group with gid 110011 was found which is not expected"


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__invalidate_group_cache_after_stop(client: Client, provider: GenericProvider):
    """
    :title: Invalidate group cache after SSSD is stopped
    :setup:
        1. Add 'group1' to SSSD
        2. Set group gid
        3. Start SSSD
    :steps:
        1. Find the 'group1' getent.group(name)
        2. Check that group has correct id
        3. Check that group has correct name
        4. Stop SSSD
        5. Invalidate cache for 'group1'
        6. Find the 'group1' getent.group(name) and getent.group(uid)
    :expectedresults:
        1. Group is found
        2. Group has correct id
        3. Group has correct name
        4. SSSD is stopped
        5. Cache is invalidated
        6. Group is not found
    :customerscenario: False
    """
    provider.group("group1").add(gid=101010)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    result = client.tools.getent.group("group1")
    assert result is not None, "Group group1 was not found using getent"
    assert result.name == "group1", f"Groupname {result.name} is incorrect, group1 expected"
    assert result.gid == 101010, f"Group gid {result.gid} is incorrect, 101010 expected"

    client.sssd.stop()
    client.sssctl.cache_expire(group="group1")

    assert client.tools.getent.group("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.getent.group(110011) is None, "Group with gid 110011 was found which is not expected"


@pytest.mark.importance("high")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__invalidate_groups_cache_before_stop(client: Client, provider: GenericProvider):
    """
    :title: Invalidate groups cache before SSSD is stopped
    :setup:
        1. Add 'group1' and 'group2' to SSSD
        2. Set groups gids
        3. Start SSSD
    :steps:
        1. Find groups with getent.group(name)
        2. Check that groups have correct gids
        3. Invalidate cache for all groups
        4. Stop SSSD
        5. Find 'group1' and 'group2' with getent.group(name) and getent.group(gid)
    :expectedresults:
        1. Groups are found
        2. Groups have correct gids
        3. Cache is invalidated
        4. SSSD is stopped
        5. Groups are not found
    :customerscenario: False
    """
    provider.group("group1").add(gid=101010)
    provider.group("group2").add(gid=202020)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    result = client.tools.getent.group("group1")
    assert result is not None, "Group group1 was not found using getent"
    assert result.gid == 101010, f"Group gid {result.gid} is incorrect, 101010 expected"

    result = client.tools.getent.group("group2")
    assert result is not None, "Group group2 was not found using getent"
    assert result.gid == 202020, f"Group gid {result.gid} is incorrect, 202020 expected"

    client.sssctl.cache_expire(groups=True)
    client.sssd.stop()

    assert client.tools.getent.group("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.getent.group(110011) is None, "Group with gid 110011 was found which is not expected"
    assert client.tools.getent.group("group2") is None, "Group group2 was found which is not expected"
    assert client.tools.getent.group(202020) is None, "Group with gid 202020 was found which is not expected"


@pytest.mark.importance("high")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__invalidate_groups_cache_after_stop(client: Client, provider: GenericProvider):
    """
    :title: Invalidate groups cache after SSSD is stopped
    :setup:
        1. Add 'group1' and 'group2' to SSSD
        2. Set groups gids
        3. Start SSSD
    :steps:
        1. Find groups with getent.group(name)
        2. Check that groups have correct gids
        3. Stop SSSD
        4. Invalidate cache for all groups
        5. Find 'group1' and 'group2' with getent.group(name) and getent.group(gid)
    :expectedresults:
        1. Groups are found
        2. Groups have correct gids
        3. SSSD is stopped
        4. Cache is invalidated
        5. Groups are not found
    :customerscenario: False
    """
    provider.group("group1").add(gid=101010)
    provider.group("group2").add(gid=202020)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    result = client.tools.getent.group("group1")
    assert result is not None, "Group group1 was not found using getent"
    assert result.gid == 101010, f"Group gid {result.gid} is incorrect, 101010 expected"

    result = client.tools.getent.group("group2")
    assert result is not None, "Group group2 was not found using getent"
    assert result.gid == 202020, f"Group gid {result.gid} is incorrect, 202020 expected"

    client.sssd.stop()
    client.sssctl.cache_expire(groups=True)

    assert client.tools.getent.group("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.getent.group(110011) is None, "Group with gid 110011 was found which is not expected"
    assert client.tools.getent.group("group2") is None, "Group group2 was found which is not expected"
    assert client.tools.getent.group(202020) is None, "Group with gid 202020 was found which is not expected"


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__invalidate_everything_before_stop(client: Client, provider: GenericProvider):
    """
    :title: Invalidate all parts of cache before SSSD is stopped
    :setup:
        1. Add 'user1' and 'user2' to SSSD
        2. Set users uids and gids
        3. Add 'group1' and 'group2' to SSSD
        4. Set groups gids
        5. Add members to the groups
        6. Start SSSD
    :steps:
        1. Find users with id(name)
        2. Check that users have correct uids
        3. Find groups with getent.group(name)
        4. Check that groups have correct gids
        5. Invalidate all parts of cache
        6. Stop SSSD
        7. Find 'user1' and 'user2' with id(name) and id(uid)
        8. Find 'group1' and 'group2' with getent.group(name) and getent.group(gid)
    :expectedresults:
        1. Users are found
        2. Users have correct uids
        3. Groups are found
        4. Groups have correct gids
        5. Cache is invalidated
        6. SSSD is stopped
        7. Users are not found
        8. Groups are not found
    :customerscenario: False
    """
    u1 = provider.user("user1").add(uid=123456, gid=110011)
    u2 = provider.user("user2").add(uid=220022, gid=222222)

    provider.group("group1").add(gid=101010).add_member(u1)
    provider.group("group2").add(gid=202020).add_members([u1, u2])

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    result = client.tools.id("user1")
    assert result is not None, "User user1 was not found using id"
    assert result.user.id == 123456, f"User id {result.user.id} is incorrect, 123456 expected"

    result = client.tools.id("user2")
    assert result is not None, "User user2 was not found using id"
    assert result.user.id == 220022, f"User id {result.user.id} is incorrect, 220022 expected"

    gresult = client.tools.getent.group("group1")
    assert gresult is not None, "Group group1 was not found using getent"
    assert gresult.gid == 101010, f"Group gid {gresult.gid} is incorrect, 101010 expected"

    gresult = client.tools.getent.group("group2")
    assert gresult is not None, "Group group2 was not found using getent"
    assert gresult.gid == 202020, f"Group gid {gresult.gid} is incorrect, 202020 expected"

    client.sssctl.cache_expire(everything=True)
    client.sssd.stop()

    assert client.tools.id("user1") is None, "User user1 was found which is not expected"
    assert client.tools.id(123456) is None, "User with id 123456 was found which is not expected"
    assert client.tools.id("user2") is None, "User user2 was found which is not expected"
    assert client.tools.id(220022) is None, "User with id 220022 was found which is not expected"
    assert client.tools.getent.group("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.getent.group(110011) is None, "Group with gid 110011 was found which is not expected"
    assert client.tools.getent.group("group2") is None, "Group group2 was found which is not expected"
    assert client.tools.getent.group(202020) is None, "Group with gid 202020 was found which is not expected"


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__invalidate_everything_after_stop(client: Client, provider: GenericProvider):
    """
    :title: Invalidate all parts of cache after SSSD is stopped
    :setup:
        1. Add 'user1' and 'user2' to SSSD
        2. Set users uids and gids
        3. Add 'group1' and 'group2' to SSSD
        4. Set groups gids
        5. Add members to the groups
        6. Start SSSD
    :steps:
        1. Find users with id(name)
        2. Check that users have correct uids
        3. Find groups with getent.group(name)
        4. Check that groups have correct gids
        5. Stop SSSD
        6. Invalidate all parts of cache
        7. Find 'user1' and 'user2' with id(name) and id(uid)
        8. Find 'group1' and 'group2' with getent.group(name) and getent.group(gid)
    :expectedresults:
        1. Users are found
        2. Users have correct uids
        3. Groups are found
        4. Groups have correct gids
        5. SSSD is stopped
        6. Cache is invalidated
        7. Users are not found
        8. Groups are not found
    :customerscenario: False
    """
    u1 = provider.user("user1").add(uid=123456, gid=110011)
    u2 = provider.user("user2").add(uid=220022, gid=222222)

    provider.group("group1").add(gid=101010).add_member(u1)
    provider.group("group2").add(gid=202020).add_members([u1, u2])

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    result = client.tools.id("user1")
    assert result is not None, "User user1 was not found using id"
    assert result.user.id == 123456, f"User id {result.user.id} is incorrect, 123456 expected"

    result = client.tools.id("user2")
    assert result is not None, "User user2 was not found using id"
    assert result.user.id == 220022, f"User id {result.user.id} is incorrect, 220022 expected"

    gresult = client.tools.getent.group("group1")
    assert gresult is not None, "Group group1 was not found using getent"
    assert gresult.gid == 101010, f"Group gid {gresult.gid} is incorrect, 101010 expected"

    gresult = client.tools.getent.group("group2")
    assert gresult is not None, "Group group2 was not found using getent"
    assert gresult.gid == 202020, f"Group gid {gresult.gid} is incorrect, 202020 expected"

    client.sssd.stop()
    client.sssctl.cache_expire(everything=True)

    assert client.tools.id("user1") is None, "User user1 was found which is not expected"
    assert client.tools.id(123456) is None, "User with id 123456 was found which is not expected"
    assert client.tools.id("user2") is None, "User user2 was found which is not expected"
    assert client.tools.id(220022) is None, "User with id 220022 was found which is not expected"
    assert client.tools.getent.group("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.getent.group(110011) is None, "Group with gid 110011 was found which is not expected"
    assert client.tools.getent.group("group2") is None, "Group group2 was found which is not expected"
    assert client.tools.getent.group(202020) is None, "Group with gid 202020 was found which is not expected"


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__memcache_timeout_zero(client: Client, provider: GenericProvider):
    """
    :title: Cache is not created at all when 'memcache_timeout' set to '0'
    :setup:
        1. Add 'user1' to SSSD
        2. Set user uid
        3. Add 'group1' to SSSD
        4. Set group gid
        5. In SSSD nss change 'memcache_timeout' set to '0'
        6. Start SSSD
    :steps:
        1. Check that cache is not created
        2. Find user with id(name)
        3. Check that user has correct uid
        4. Find group with getent.group(name)
        5. Check that group has correct gid
        6. Stop SSSD
        7. Find user with id(name) and id(uid)
        8. Find group with getent.group(name) and getent.group(gid)
    :expectedresults:
        1. Cache is not created
        2. User is found
        3. User has correct uid
        4. Group is found
        5. Group has correct gid
        6. Stop SSSD
        7. User is not found
        8. Group is not found
    :customerscenario: False
    """
    provider.user("user1").add(uid=123456, gid=19001)
    provider.group("group1").add(gid=10001)

    client.sssd.nss["memcache_timeout"] = "0"
    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    r = client.host.conn.exec(["ls", "/var/lib/sss/mc"])
    assert r.stdout == "", "Cache directory is not empty"
    assert r.stderr == "", "Ls command failed"

    result = client.tools.id("user1")
    assert result is not None, "User user1 was not found using user1"
    assert result.user.id == 123456, f"User id {result.user.id} is incorrect, 123456 expected"

    gresult = client.tools.getent.group("group1")
    assert gresult is not None, "Group group1 is not found using getent"
    assert gresult.gid == 10001, f"Group gid {gresult.gid} is incorrect, 10001 expected"

    client.sssd.stop()

    assert client.tools.id("user1") is None, "User user1 was found which is not expected"
    assert client.tools.id(123456) is None, "User with id 123456 was found which is not expected"
    assert client.tools.getent.group("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.getent.group(10001) is None, "Group with gid 10001 was found which is not expected"


@pytest.mark.importance("critical")
@pytest.mark.cache
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_memcache__removed_cache_without_invalidation(client: Client, provider: GenericProvider):
    """
    :title: SSSD is stopped, cache removed then users and groups cannot be lookedup
    :setup:
        1. Add 'user1' to SSSD
        2. Set user uid and gid
        3. Add 'group1' to SSSD
        4. Set group gid
        5. Start SSSD
    :steps:
        1. Find user with id(name)
        2. Check that user has correct uid
        3. Find group with getent.group(name)
        4. Check that group has correct gid
        5. Stop SSSD
        6. Remove cache files
        7. Find user with id(name) and id(uid)
        8. Find group with getent.group(name) and getent.group(gid)
    :expectedresults:
        1. User is found
        2. User has correct uid
        3. Group is found
        4. Group has correct gid
        5. SSSD is stopped
        6. Cache files are removed
        7. User is not found
        8. Group is not found
    :customerscenario: True
    """
    provider.user("user1").add(uid=123456, gid=19001)
    provider.group("group1").add(gid=10001)

    client.sssd.domain["ldap_id_mapping"] = "false"
    client.sssd.start()

    result = client.tools.id("user1")
    assert result is not None, "User user1 is not found using id"
    assert result.user.id == 123456, f"User id {result.user.id} is incorrect, 123456 expected"

    gresult = client.tools.getent.group("group1")
    assert gresult is not None, "Group group1 is not found using getent"
    assert gresult.gid == 10001, f"Group gid {gresult.gid} is incorrect, 10001 expected"

    client.sssd.stop()

    r = client.host.conn.exec(["ls", "/var/lib/sss/mc"])
    for file in r.stdout.split():
        check = client.host.conn.exec(["rm", f"/var/lib/sss/mc/{file}"])
        assert check.rc == 0, "Cache file was not removed successfully"

    assert client.tools.id("user1") is None, "User user1 was found which is not expected"
    assert client.tools.id(123456) is None, "User with id 123456 was found which is not expected"
    assert client.tools.getent.group("group1") is None, "Group group1 was found which is not expected"
    assert client.tools.getent.group(10001) is None, "Group with gid 10001 was found which is not expected"


@pytest.mark.topology(KnownTopology.LDAP)
@pytest.mark.ticket(bz=2226021)
def test_memcache__truncate_in_memory_cache_no_sigbus(client: Client, ldap: LDAP):
    """
    :title: Accessing truncated in-memory cache file does not cause SIGBUS
    :setup:
        1. Add 'user-1' to SSSD
        2. Start SSSD
    :steps:
        1. Find 'user-1' so it is stored in in-memory cache
        2. Truncate /var/lib/sss/mc/passwd
        3. Check that 'user-1' can be correctly resolved
    :expectedresults:
        1. User is found
        2. Size of /var/lib/sss/mc/passwd is 0
        3. User can be found again and there is no crash
    :customerscenario: True
    """
    ldap.user("user-1").add()

    client.sssd.start()

    result = client.tools.id("user-1")
    assert result is not None, "User was not found"
    assert result.user.name == "user-1"

    client.fs.truncate("/var/lib/sss/mc/passwd")

    result = client.tools.id("user-1")
    assert result is not None, "User was not found"
    assert result.user.name == "user-1"



import re
import pytest

from sssd.testlib.common.utils import sssdTools


@pytest.mark.tier1
@pytest.mark.admultidomain
class TestADMultiDomain(object):

    @staticmethod
    def test_0001_bz2013297(multihost, newhostname, adchildjoin):
        """
        :title: IDM-SSSD-TC: ad_provider: forests: disabled root domain causes subdomains to be marked offline
        :id: 3055d093-8449-4146-a6e1-b221dee35395
        :setup:
          1. Configure parent and child domain
          2. Join client to child domain
          3. ad_enabled_domains is not configured
          4. ad_enabled_domains to contain only the child domain
        :steps:
          1. Lookup user from child domain
          2. Lookup user from parent domain
          3. Change  ad_enabled_domains parameter
          4. Lookup user from child domain
          5. Lookup user from parent domain
        :expectedresults:
          1. Parent user is found
          2. Child user is found
          3. Parent user is not found
          4. Child user is found
        :customerscenario: True
        """
        adchildjoin(membersw='adcli')
        ad_domain = multihost.ad[0].domainname
        child_domain = multihost.ad[1].domainname
        ad_server = multihost.ad[1].hostname

        multihost.client[0].service_sssd('stop')
        client = sssdTools(multihost.client[0], multihost.ad[1])
        client.backup_sssd_conf()
        dom_section = f'domain/{client.get_domain_section_name()}'
        sssd_params = {
            'ad_domain': child_domain,
            'debug_level': '9',
            'use_fully_qualified_names': 'True',
            'ad_server': ad_server,
            'cache_credentials': 'True',
        }
        client.sssd_conf(dom_section, sssd_params, action='update')
        client.clear_sssd_cache()
        multihost.client[0].service_sssd('start')

        getent_root_user1 = multihost.client[0].run_command(
            f'getent passwd user1@{ad_domain}', raiseonerr=False)
        getent_child_user1 = multihost.client[0].run_command(
            f'getent passwd child_user1@{child_domain}', raiseonerr=False)

        client.restore_sssd_conf()
        client.clear_sssd_cache()

        assert getent_root_user1.returncode == 0
        assert getent_child_user1.returncode == 0

        dom_section = f'domain/{client.get_domain_section_name()}'
        sssd_params = {
            'ad_domain': child_domain,
            'debug_level': '9',
            'use_fully_qualified_names': 'True',
            'cache_credentials': 'True',
            'ad_server': ad_server,
            'ad_enabled_domains': child_domain
        }
        client.sssd_conf(dom_section, sssd_params, action='update')
        client.clear_sssd_cache()
        multihost.client[0].service_sssd('start')

        getent_root_user2 = multihost.client[0].run_command(
            f'getent passwd user1@{ad_domain}', raiseonerr=False)
        getent_child_user2 = multihost.client[0].run_command(
            f'getent passwd child_user1@{child_domain}', raiseonerr=False)

        client.restore_sssd_conf()
        client.clear_sssd_cache()

        assert getent_root_user2.returncode == 2
        assert getent_child_user2.returncode == 0

    @staticmethod
    def test_0002_bz2018432(multihost, newhostname, adjoin):
        """
        :title: IDM-SSSD-TC: ad_provider: forests: sssctl domain_list shows more domains than it should
        :id: b2c9efc8-b3a6-4216-99d6-7ae1d868c43f
        :setup:
          1. Configure several domains, this suite contains 4 trusted domains
          2. Join client to parent domain
        :steps:
          1. Perform sssctl domain-list
        :expectedresults:
          1. Only trusted domains listed
        :customerscenario: True
        """
        adjoin(membersw='adcli')
        ad_domain = multihost.ad[0].domainname
        ad_server = multihost.ad[0].hostname
        multihost.client[0].service_sssd('stop')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        client.backup_sssd_conf()
        dom_section = f'domain/{client.get_domain_section_name()}'
        sssd_params = {
            'ad_domain': ad_domain,
            'debug_level': '9',
            'use_fully_qualified_names': 'True',
            'ad_server': ad_server,
            'cache_credentials': 'True'
        }
        client.sssd_conf(dom_section, sssd_params, action='update')
        client.clear_sssd_cache()
        multihost.client[0].service_sssd('start')
        # The output needs to be pruned of servers that are not apart of the forest and 'implicit files'
        domain_list_cmd = multihost.client[0].run_command('sssctl domain-list', raiseonerr=False)
        domain_list = domain_list_cmd.stdout_text.split('\n')
        if "" in domain_list:
            domain_list.remove("")
        if "implicit_files" in domain_list:
            domain_list.remove("implicit_files")
        multihost_list = []
        for x in multihost.ad:
            multihost_list.append(x.domainname)
        multihost_list.pop()

        domain_list.sort()
        multihost_list.sort()

        assert domain_list == multihost_list

    @staticmethod
    def test_0003_bz2167728(multihost, newhostname, adchildjoin):
        """
        :title: IDM-SSSD-TC: ad_provider: forests: bz2167728 Auth fails if client cannot speak to forest root domain
        :id: e9ba9423-0a42-4379-a900-637c79ff0e5c
        :setup:
          1. Clear out the contents of [domain_realm] in /etc/krb5.conf
          2. Join client to child domain
        :steps:
          1. Lookup root, child and tree domain users
        :expectedresults:
          1. All lookups should work
        :customerscenario: True
        """
        client = sssdTools(multihost.client[0], multihost.ad[1])
        krb5 = multihost.client[0].get_file_contents('/etc/krb5.conf', encoding='utf-8')
        resolv = multihost.client[0].get_file_contents('/etc/resolv.conf', encoding='utf-8')
        domain = multihost.ad[0].domainname
        ip = multihost.ad[0].ip
        child_domain = multihost.ad[1].domainname
        child_ip = multihost.ad[1].ip
        tree_domain = multihost.ad[2].domainname
        tree_ip = multihost.ad[2].ip

        # To verify this bug the contents of /etc/krb5.conf needs to have no [realm] entries
        for x in multihost.ad:
            _domain = x.domainname
            _domain_upper = _domain.capitalize()
            _krb5 = multihost.client[0].get_file_contents('/etc/krb5.conf', encoding='utf-8')
            _krb5_1 = re.sub(f"^.{_domain} = {_domain_upper}", "", re.sub(f"^{_domain} = {_domain_upper}", "", _krb5))
            multihost.client[0].put_file_contents('/etc/krb5.conf', _krb5_1)
        adchildjoin(membersw='adcli')

        multihost.client[0].service_sssd('stop')
        client.backup_sssd_conf()
        sssd_domain = f'domain/{client.get_domain_section_name()}'
        sssd_params = {'debug_level': '9'}
        client.sssd_conf(sssd_domain, sssd_params, action='update')

        client.update_resolv_conf(child_ip)
        client.update_resolv_conf(ip)
        client.update_resolv_conf(tree_ip)

        multihost.client[0].service_sssd('start')

        getent1 = multihost.client[0].run_command(f'getent passwd user1@{domain}', raiseonerr=False)
        getent2 = multihost.client[0].run_command(f'getent passwd child_user1@{child_domain}', raiseonerr=False)
        getent3 = multihost.client[0].run_command(f'getent passwd tree_user1@{tree_domain}', raiseonerr=False)

        multihost.client[0].put_file_contents('/etc/krb5.conf', krb5)
        multihost.client[0].run_command('chattr -i /etc/resolv.conf', raiseonerr=False)
        multihost.client[0].put_file_contents('/etc/resolv.conf', resolv)
        multihost.client[0].run_command('chattr +i /etc/resolv.conf', raiseonerr=False)
        client.restore_sssd_conf()

        assert getent1.returncode == 0, f'Could not find user1@{domain}!'
        assert getent2.returncode == 0, f'Could not find child_user1@{child_domain}!'
        assert getent3.returncode == 0, f'Could not find tree_user1@{tree_domain}!'

    @pytest.mark.ticket(bz=1913284, jira=["SSSD-3092", "RHEL-4974"])
    @staticmethod
    def test_0004_bz1913284_keytab_as_nonroot(multihost, newhostname, adchildjoin):
        """
        :title: IDM-SSSD-TC: ad_provider: forests: krb5_kt_start_seq_get failed:
          Permission denied when running as unprivileged user sssd
        :id: 53a6871e-95a6-4865-9e61-1e12815ec35b
        :setup:
          1. Configure parent and child domain
          2. Join client to child domain
        :steps:
          1. Lookup user from child domain
          2. Lookup user from parent domain
          3. Check log for a keytab error
        :expectedresults:
          1. Parent user is found
          2. Child user is found
          3. The permission denied error is not present in log.
        :customerscenario: True
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1913284
        """
        adchildjoin(membersw='adcli')
        ad_domain = multihost.ad[0].domainname
        child_domain = multihost.ad[1].domainname
        ad_server = multihost.ad[1].hostname

        # Configure sssd
        client = sssdTools(multihost.client[0], multihost.ad[1])
        if client.sssd_user != "sssd":
            pytest.skip("The test is not applicable without non-root fetaure (sssd 2.10+)!")
        multihost.client[0].service_sssd('stop')
        client.backup_sssd_conf()
        dom_section = f'domain/{client.get_domain_section_name()}'
        sssd_params = {
            'ad_domain': child_domain,
            'debug_level': '9',
            'use_fully_qualified_names': 'True',
            'ad_server': ad_server,
            'cache_credentials': 'True',
        }
        client.sssd_conf(dom_section, sssd_params)
        client.clear_sssd_cache()

        # Search for the user in root domain
        getent_root_user1 = multihost.client[0].run_command(
            f'getent passwd user1@{ad_domain}',
            raiseonerr=False
        )
        # Search for the user in child domain
        getent_child_user1 = multihost.client[0].run_command(
            f'getent passwd child_user1@{child_domain}',
            raiseonerr=False
        )
        # Download sssd log
        log_str = multihost.client[0].get_file_contents(
            f"/var/log/sssd/sssd_{multihost.ad[1].domainname.lower()}.log"). \
            decode('utf-8')
        client.restore_sssd_conf()
        client.clear_sssd_cache()

        # Evaluate test results
        assert getent_root_user1.returncode == 0
        assert getent_child_user1.returncode == 0
        assert "krb5_kt_start_seq_get failed: Permission denied" not in log_str
        assert "Failed to read keytab [FILE:/etc/krb5.keytab]: No " \
               "suitable principal found in keytab" not in log_str



#!/usr/bin/env python
#  SSSD
#
#  Unit tests for pysss_murmur
#
#  Copyright (C) Sumit Bose <sbose@redhat.com>        2012
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import print_function

import unittest
import sys
import os
import tempfile

BUILD_DIR = os.getenv('builddir') or "."
TEST_DIR = os.path.realpath(os.getenv('SSS_TEST_DIR') or ".")
MODPATH = tempfile.mkdtemp(prefix="tp_pysss_murmur_", dir=TEST_DIR)


class PySssMurmurImport(unittest.TestCase):
    def setUp(self):
        " Make sure we load the in-tree module "
        self.system_path = sys.path[:]
        sys.path = [MODPATH]
        print(os.getcwd())
        print(MODPATH)

    def tearDown(self):
        " Restore the system path "
        sys.path = self.system_path

    def testImport(self):
        " Import the module and assert it comes from tree "
        try:
            dest_module_path = MODPATH + "/pysss_murmur.so"

            if sys.version_info[0] > 2:
                src_module_path = BUILD_DIR + "/.libs/_py3sss_murmur.so"
            else:
                src_module_path = BUILD_DIR + "/.libs/_py2sss_murmur.so"

            src_module_path = os.path.abspath(src_module_path)
            os.symlink(src_module_path, dest_module_path)

            import pysss_murmur
        except ImportError as e:
            print("Could not load the pysss_murmur module. "
                  "Please check if it is compiled", file=sys.stderr)
            raise e
        self.assertEqual(os.path.realpath(pysss_murmur.__file__),
                         os.path.realpath(MODPATH + "/pysss_murmur.so"))


class PySssMurmurTestNeg(unittest.TestCase):
    def test_invalid_arguments(self):
        self.assertRaises(ValueError, pysss_murmur.murmurhash3, 1, 2, 3)
        self.assertRaises(ValueError, pysss_murmur.murmurhash3, "test", 2)
        self.assertRaises(ValueError, pysss_murmur.murmurhash3, "test")
        self.assertRaises(ValueError, pysss_murmur.murmurhash3)

    def test_invalid_length(self):
        seed = 12345

        self.assertRaises(ValueError, pysss_murmur.murmurhash3, "t", -1, seed)
        # length is off by one
        self.assertRaises(ValueError, pysss_murmur.murmurhash3, "test", 5,
                          seed)
        self.assertRaises(ValueError, pysss_murmur.murmurhash3, "test",
                          0xffffffffff, seed)


class PySssMurmurTestPos(unittest.TestCase):
    @classmethod
    def tearDownClass(cls):
        os.unlink(MODPATH + "/pysss_murmur.so")
        os.rmdir(MODPATH)

    def testExpectedHash(self):
        sid_str = "S-1-5-21-2153326666-2176343378-3404031434"
        seed = 0xdeadbeef

        hash_val = pysss_murmur.murmurhash3(sid_str, 0, seed)
        self.assertEqual(hash_val, 233162409)

        hash_val = pysss_murmur.murmurhash3(sid_str, len(sid_str), seed)
        self.assertEqual(hash_val, 93103853)

    def test_memory_cache_usage(self):
        seed = 0xbeefdead
        input_str = "test_user1"
        input_len = len(input_str)

        val_bin = pysss_murmur.murmurhash3(input_str + '\0',
                                           input_len + 1, seed)
        self.assertEqual(val_bin, 1198610880)

        val_bin = pysss_murmur.murmurhash3(input_str + '\0' * 5,
                                           input_len + 5, seed)
        self.assertEqual(val_bin, 2917868047)


if __name__ == "__main__":
    error = 0

    suite = unittest.TestLoader().loadTestsFromTestCase(PySssMurmurImport)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x1
        # need to bail out here because pysss_murmur could not be imported
        sys.exit(error)

    # import the pysss_murmur module into the global namespace, but make sure
    # it's the one in tree
    sys.path.insert(0, MODPATH)
    import pysss_murmur

    suite = unittest.TestLoader().loadTestsFromTestCase(PySssMurmurTestNeg)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x2

    suite = unittest.TestLoader().loadTestsFromTestCase(PySssMurmurTestPos)
    res = unittest.TextTestRunner().run(suite)
    if not res.wasSuccessful():
        error |= 0x4

    sys.exit(error)



"""
Passkey Tests.

:requirement: passkey

Passkeys allow users to authenticate without having to enter a username or password,
or provide any additional authentication factor.
This technology aims to replace legacy authentication mechanisms such as passwords.

The objective is to use a passkey to locally authenticate a user against centralized identity management system.
For that purpose an integration with a server like AD/Samba or IPA or LDAP is needed.

The passkey solution only enables to authenticate in a system where the
FIDO2 key is connected physically.This could be su login, the GDM login.
The passkey is another way to authenticate as the user, using a physical token.

We can't support remote authentication (ssh) because there isn't any way of doing the remote authentication
when the key is attached to your laptop.
Here, passkey support is tested with su, tests are running with
umockdev, not with a physical key.

We are creating the recording files and reusing them in test without having passkey connected to host.
To create the recording files we have to connect passkey and need biometric
authentication such as pin and finger touch.

we use sssctl tool to create the passkey-mapping.
# sssctl passkey-register --username=<username> --domain=<domain name>
Next, it will ask for PIN and generate the passkey-mapping and token.

.. code-block::
    mapping = client.sssctl.passkey_register(
        username="user1",
        domain="ldap.test",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script",
    )

Once we add user along with passkey-mapping, we can test/assert the passkey authentication.
While authenticating we need to username, pin of passkey and some recording files which will use for authenticating
the user.

.. code-block::
    assert client.auth.su.passkey(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.{suffix}",
    )

For IPA tests where we need to test commands after authentication of user, we the use following code.
Here, we have an extra argument as a command to test in session after authentication of user.
It returns returncode either 0 or 1 and output to fetch the console messages.

.. code-block::
    rc, _, output, _ = client.auth.su.passkey_with_output(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.ipa",
        command="klist",
    )
"""

from __future__ import annotations

import re

import pytest
from pytest_mh import mh_fixture
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.generic import GenericProvider
from sssd_test_framework.roles.ipa import IPA
from sssd_test_framework.topology import KnownTopology, KnownTopologyGroup
from sssd_test_framework.utils.authentication import PasskeyAuthenticationUseCases


@mh_fixture()
def umockdev_ipaotpd_update(ipa: IPA, request: pytest.FixtureRequest):
    """
    Update the ipa-optd@.service file from ipa server with
    'Environment=LD_PRELOAD=/opt/random.so' to avoid the data mismatch
    error while running the umockdev-run command while authenticating the user.
    """
    ipa.fs.append("/usr/lib/systemd/system/ipa-otpd@.service", "Environment=LD_PRELOAD=/opt/random.so")
    ipa.svc.restart("ipa")


@pytest.mark.importance("high")
@pytest.mark.topology(KnownTopology.Client)
@pytest.mark.builtwith(client="passkey")
def test_passkey__register_sssctl(client: Client, moduledatadir: str, testdatadir: str):
    """
    :title: Register a key with sssctl
    :setup:
        1. Setup IDM client with FIDO and umockdev setup
    :steps:
        1. Use sssctl to register a FIDO2 key.
        2. Check the output.
    :expectedresults:
        1. New key is registered.
        2. Output contains key mapping data.
    :customerscenario: False
    """
    mapping = client.sssctl.passkey_register(
        username="user1",
        domain="ldap.test",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script",
    )

    with open(f"{testdatadir}/passkey-mapping") as f:
        assert mapping == f.read().strip(), "Failed to register a key with sssctl"


@pytest.mark.importance("high")
@pytest.mark.topology(KnownTopology.IPA)
@pytest.mark.builtwith(client="passkey", ipa="passkey")
def test_passkey__register_ipa(ipa: IPA, moduledatadir: str, testdatadir: str):
    """
    :title: Register a passkey with the IPA command
    :setup:
        1. Setup IDM client with FIDO and umockdev setup
    :steps:
        1. Use ipa command to register a FIDO2 key.
        2. Check the output that contains the user key mapping data.
    :expectedresults:
        1. New key is registered with IPA command.
        2. Output contains key mapping data.
    :customerscenario: False
    """
    mapping = (
        ipa.user("user1")
        .add()
        .passkey_add_register(
            pin=123456,
            device=f"{moduledatadir}/umockdev.device",
            ioctl=f"{moduledatadir}/umockdev.ioctl",
            script=f"{testdatadir}/umockdev.script",
        )
    )

    with open(f"{testdatadir}/passkey-mapping") as f:
        assert mapping == f.read().strip(), "Failed to register a key with the IPA command"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.builtwith(client="passkey", provider="passkey")
def test_passkey__su_user(client: Client, provider: GenericProvider, moduledatadir: str, testdatadir: str):
    """
    :title: Check su authentication of user with LDAP, IPA, AD and Samba
    :setup:
        1. Add a user in LDAP, IPA, AD and Samba with passkey_mapping.
        2. Setup SSSD client with FIDO and umockdev, start SSSD service.
    :steps:
        1. Check su authentication of the user.
    :expectedresults:
        1. User su authenticates successfully.
    :customerscenario: False
    """
    suffix = type(provider).__name__.lower()

    client.sssd.domain["local_auth_policy"] = "only"

    with open(f"{testdatadir}/passkey-mapping.{suffix}") as f:
        provider.user("user1").add().passkey_add(f.read().strip())

    client.sssd.start(service_user="root")

    assert client.auth.su.passkey(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.{suffix}",
    )


@pytest.mark.importance("high")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.builtwith(client="passkey", provider="passkey")
def test_passkey__su_user_with_failed_pin(
    client: Client, provider: GenericProvider, moduledatadir: str, testdatadir: str
):
    """
    :title: Check su authentication deny of user with LDAP, IPA, AD and Samba with incorrect pin
    :setup:
        1. Add a LDAP, IPA, AD and Samba user with passkey_mapping.
        2. Setup SSSD client with FIDO and umockdev, start SSSD service.
    :steps:
        1. Check su authentication of the user with incorrect PIN.
    :expectedresults:
        1. User failed to su authenticate.
    :customerscenario: False
    """
    suffix = type(provider).__name__.lower()

    client.sssd.domain["local_auth_policy"] = "only"

    with open(f"{testdatadir}/passkey-mapping.{suffix}") as f:
        provider.user("user1").add().passkey_add(f.read().strip())

    client.sssd.start(service_user="root")

    assert not client.auth.su.passkey(
        username="user1",
        pin=67890,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.{suffix}",
    )


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.builtwith(client="passkey", provider="passkey")
def test_passkey__su_user_with_incorrect_mapping(
    client: Client, provider: GenericProvider, moduledatadir: str, testdatadir: str
):
    """
    :title: Check su authentication deny of user with LDAP, IPA, AD and Samba with incorrect mapping
    :setup:
        1. Add a LDAP, IPA, AD and Samba user with passkey_mapping.
        2. Setup SSSD client with FIDO and umockdev, start SSSD service.
    :steps:
        1. Check su authentication of the user with incorrect passkey mapping.
    :expectedresults:
        1. User failed to su authenticate.
    :customerscenario: False
    """
    suffix = type(provider).__name__.lower()

    client.sssd.domain["local_auth_policy"] = "only"

    # Here, we are using passkey-mapping from the other FIDO2 key.

    with open(f"{testdatadir}/passkey-mapping.{suffix}") as f:
        provider.user("user1").add().passkey_add(f.read().strip())

    client.sssd.start(service_user="root")

    assert not client.auth.su.passkey(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.{suffix}",
    )


@pytest.mark.importance("high")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.builtwith(client="passkey", provider="passkey")
def test_passkey__su_user_when_server_is_not_resolvable(
    client: Client, provider: GenericProvider, moduledatadir: str, testdatadir: str
):
    """
    :title: Check su authentication of a user with LDAP, IPA, AD and Samba when server is not resolvable
    :setup:
        1. Add a LDAP, IPA, AD and Samba user with passkey_mapping.
        2. Setup SSSD client with FIDO and umockdev, start SSSD service.
    :steps:
        1. Check su authentication of the user.
        2. Update the server url and restart the sssd service to reflect the changes.
        3. Check su authentication of the user.
    :expectedresults:
        1. User su authenticates successfully.
        2. Successfully update the sssd.conf and restarted the sssd service.
        3. User su authenticates successfully due to cached data.
    :customerscenario: False
    """
    suffix = type(provider).__name__.lower()
    if suffix == "ipa":
        server_url = "ipa_server"
    elif suffix == "ldap":
        server_url = "ldap_uri"
    elif suffix == "samba" or "ad":
        server_url = "ad_server"
    else:
        assert False, "provider not found"

    client.sssd.domain["local_auth_policy"] = "only"

    with open(f"{testdatadir}/passkey-mapping.{suffix}") as f:
        provider.user("user1").add().passkey_add(f.read().strip())

    client.sssd.start(service_user="root")

    # First time check authentication to cache the user
    assert client.auth.su.passkey(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.{suffix}",
    )

    # Here we are making server/backend offline but not deleting cache and logs.
    client.sssd.config.remove_option("domain/test", server_url)
    client.sssd.domain[server_url] = "ldap://new.server.test"
    client.sssd.start()

    assert client.auth.su.passkey(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.{suffix}",
    )


@pytest.mark.importance("high")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.builtwith(client="passkey", provider="passkey")
def test_passkey__su_user_when_offline(
    client: Client, provider: GenericProvider, moduledatadir: str, testdatadir: str
):
    """
    :title: Check offline su authentication of a user with LDAP, IPA, AD and Samba
    :setup:
        1. Add a LDAP, IPA, AD and Samba user with passkey_mapping.
        2. Setup SSSD client with FIDO and umockdev, start SSSD service.
    :steps:
        1. Check su authentication of the user.
        2. Make server offline (by blocking traffic to the provider).
        3. Bring SSSD offline explicitly.
        4. Check su offline authentication of the user.
    :expectedresults:
        1. User su authenticated successfully.
        2. Firewall rule added, traffic is dropped.
        3. SSSD is offline.
        4. Offline su authentication is successful.
    :customerscenario: False
    """
    suffix = type(provider).__name__.lower()

    with open(f"{testdatadir}/passkey-mapping.{suffix}") as f:
        provider.user("user1").add().passkey_add(f.read().strip())

    client.sssd.domain["local_auth_policy"] = "only"

    client.sssd.start(service_user="root")

    # First time check authentication to cache the user
    assert client.auth.su.passkey(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.{suffix}",
    )

    # Render the provider offline
    client.firewall.outbound.reject_host(provider)

    # There might be active connections that are not terminated by creating firewall rule.
    # We need to terminated it by bringing SSSD to offline state explicitly.
    client.sssd.bring_offline()

    assert client.auth.su.passkey(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.{suffix}",
    )


@pytest.mark.importance("high")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.builtwith(client="passkey", provider="passkey")
def test_passkey__lookup_user_from_cache(
    client: Client, provider: GenericProvider, moduledatadir: str, testdatadir: str
):
    """
    :title: Fetch a user from cache for LDAP, IPA, AD and Samba server
    :setup:
        1. Add a user in LDAP, IPA, AD and Samba with passkey_mapping.
        2. Setup SSSD client with FIDO and umockdev, start SSSD service.
    :steps:
        1. Check a user lookup.
        2. Check a user from cache using ldbsearch command.
    :expectedresults:
        1. A user looked up successfully.
        2. Successfully get the user from ldbsearch command.
    :customerscenario: False
    """

    suffix = type(provider).__name__.lower()

    with open(f"{testdatadir}/passkey-mapping.{suffix}") as f:
        provider.user("user1").add().passkey_add(f.read().strip())

    client.sssd.start()

    result = client.tools.id("user1")
    output = client.ldb.search(
        path="/var/lib/sss/db/cache_test.ldb", basedn="name=user1@test,cn=users,cn=test,cn=sysdb", filter="userPasskey"
    )
    assert result is not None
    assert output["name=user1@test,cn=users,cn=test,cn=sysdb"] is not None
    assert "name=user1@test,cn=users,cn=test,cn=sysdb" in output.keys(), "user not find in cache"
    assert "userPasskey" in (list(output.values())[0].keys()), "passkey mapping is not found in cache"


@pytest.mark.importance("high")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.builtwith(client="passkey", provider="passkey")
def test_passkey__su_user_with_multiple_keys(
    client: Client, provider: GenericProvider, moduledatadir: str, testdatadir: str
):
    """
    :title: Check su authentication of user when multiple keys added for same user with
            LDAP, IPA, AD and Samba server.
    :setup:
        1. Add a user with multiple mappings of passkey in LDAP, IPA, AD and Samba with passkey_mapping.
        2. Setup SSSD client with FIDO and umockdev, start SSSD service.
    :steps:
        1. Check su authentication of the user.
    :expectedresults:
        1. User su authenticates successfully.
    :customerscenario: False
    """
    suffix = type(provider).__name__.lower()
    user_add = provider.user("user1").add()

    client.sssd.domain["local_auth_policy"] = "only"

    for n in range(1, 5):
        with open(f"{testdatadir}/passkey-mapping.{suffix}{n}") as f:
            user_add.passkey_add(f.read().strip())

    client.sssd.start(service_user="root")

    assert client.auth.su.passkey(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.{suffix}",
    )


@pytest.mark.importance("high")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.builtwith(client="passkey", provider="passkey")
def test_passkey__su_user_same_key_for_other_users(
    client: Client, provider: GenericProvider, moduledatadir: str, testdatadir: str
):
    """
    :title: Check su authentication of user when same key added for multiple user with LDAP, IPA, AD and Samba server.
    :setup:
        1. Add three users with same passkey mapping in LDAP, IPA, AD and Samba with passkey_mapping.
        2. Setup SSSD client with FIDO and umockdev, start SSSD service.
    :steps:
        1. Check su authentication of the user1, user2 and user3.
    :expectedresults:
        1. User1, user2 and user3 su authenticates successfully with same mapping.
    :customerscenario: False
    """
    suffix = type(provider).__name__.lower()

    client.sssd.domain["local_auth_policy"] = "only"

    client.sssd.start(service_user="root")

    for user in ["user1", "user2", "user3"]:
        user_add = provider.user(user).add()
        with open(f"{testdatadir}/passkey-mapping.{suffix}") as f:
            user_add.passkey_add(f.read().strip())

        assert client.auth.su.passkey(
            username=user,
            pin=123456,
            device=f"{moduledatadir}/umockdev.device",
            ioctl=f"{moduledatadir}/umockdev.ioctl",
            script=f"{testdatadir}/umockdev.script.{suffix}.{user}",
        )


@pytest.mark.importance("high")
@pytest.mark.ticket(jira="SSSD-7011", gh=7066)
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
@pytest.mark.topology(KnownTopology.LDAP)
@pytest.mark.builtwith(client="passkey", provider="passkey")
def test_passkey__check_passkey_mapping_token_as_ssh_key_only(
    client: Client, provider: GenericProvider, moduledatadir: str, testdatadir: str
):
    """
    :title: Check passkey mapping with invalid ssh key with AD, Samba, and LDAP server.
    :setup:
        1. Add a users in AD, Samba and LDAP server and add ssh key as a passkey mapping.
        2. Setup SSSD client with FIDO, start SSSD service.
    :steps:
        1. Check su non-passkey authentication of the user.
        2. Required error message in pam log.
    :expectedresults:
        1. su authenticates the user with correct password.
        2. Get the expected message in pam log.
    :customerscenario: False
    """
    client.sssd.domain["local_auth_policy"] = "enable:passkey"

    with open(f"{testdatadir}/ssh-key") as f:
        provider.user("user1").add().passkey_add(f.read().strip())

    client.sssd.start()

    # We are running simple su not to check authentication with passkey but just to get
    # expected log message.
    assert client.auth.su.password("user1", "Secret123"), "Password authentication with correct password is failed"

    pam_log = client.fs.read(client.sssd.logs.pam)
    assert "Mapping data found is not passkey related" in pam_log, "String was not found in the logs"


@pytest.mark.importance("high")
@pytest.mark.ticket(jira="SSSD-7011", gh=7066)
@pytest.mark.topology(KnownTopologyGroup.AnyAD)
@pytest.mark.topology(KnownTopology.LDAP)
@pytest.mark.builtwith(client="passkey", provider="passkey")
def test_passkey__su_user_when_add_with_ssh_key_and_mapping(
    client: Client, provider: GenericProvider, moduledatadir: str, testdatadir: str
):
    """
    :title: Check authentication of user when ssh key and valid passkey mapping added with AD, Samba, and LDAP server.
    :setup:
        1. Add a users in AD, Samba and LDAP server and add ssh key and a passkey mapping.
        2. Setup SSSD client with FIDO, start SSSD service.
    :steps:
        1. Check su passkey authentication of the user.
        2. Required error message in pam log.
    :expectedresults:
        1. su authenticates the user successfully.
        2. Get the expected message in pam log.
    :customerscenario: False
    """
    suffix = type(provider).__name__.lower()

    client.sssd.domain["local_auth_policy"] = "enable:passkey"

    user_add = provider.user("user1").add()
    for mapping in ["ssh-key", f"passkey-mapping.{suffix}"]:
        with open(f"{testdatadir}/{mapping}") as f:
            user_add.passkey_add(f.read().strip())

    client.sssd.start(service_user="root")

    assert client.auth.su.passkey(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.{suffix}",
    )

    pam_log = client.fs.read(client.sssd.logs.pam)
    assert "Mapping data found is not passkey related" in pam_log, "String was not found in the logs"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
@pytest.mark.builtwith(client="passkey", provider="passkey")
def test_passkey__su_fips_fido_key(client: Client, provider: GenericProvider, moduledatadir: str, testdatadir: str):
    """
    :title: Check su authentication of user with LDAP, IPA, AD and Samba with FIPS Fido key
    :setup:
        1. Add a user in LDAP, IPA, AD and Samba with passkey_mapping.
        2. Setup SSSD client with FIDO and umockdev, start SSSD service.
    :steps:
        1. Check su authentication of the user.
    :expectedresults:
        1. User su authenticates successfully.
    :customerscenario: False
    """
    suffix = type(provider).__name__.lower()

    client.sssd.domain["local_auth_policy"] = "enable:passkey"

    # Recording files are created in FIPS enabled host with
    # FIPS Fido key.

    with open(f"{testdatadir}/passkey-mapping.{suffix}") as f:
        provider.user("user1").add().passkey_add(f.read().strip())

    client.sssd.start(service_user="root")

    assert client.auth.su.passkey(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.{suffix}",
    )


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopology.IPA)
@pytest.mark.builtwith(client="passkey", ipa="passkey")
def test_passkey__check_tgt(client: Client, ipa: IPA, moduledatadir: str, testdatadir: str, umockdev_ipaotpd_update):
    """
    :title: Check the TGT of user after authentication.
    :setup:
        1. Add a user with --user-auth-type=passkey in the server with passkey mapping.
        2. Setup SSSD client with FIDO and umockdev, start SSSD service.
    :steps:
        1. Check authentication of the user
        2. Check TGT after authenticates.
    :expectedresults:
        1. User authenticates successfully.
        2. Gets the TGT.
    :customerscenario: False
    """
    with open(f"{testdatadir}/passkey-mapping.ipa") as f:
        ipa.user("user1").add(user_auth_type="passkey").passkey_add(f.read().strip())

    client.sssd.start(service_user="root")

    rc, _, output, _ = client.auth.su.passkey_with_output(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.ipa",
        command="klist",
    )

    assert rc == 0, "Authentication failed"
    assert "Ticket cache" in output, "Failed to get the TGT"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopology.IPA)
@pytest.mark.builtwith(client="passkey", ipa="passkey")
def test_passkey__ipa_server_offline(
    client: Client, ipa: IPA, moduledatadir: str, testdatadir: str, umockdev_ipaotpd_update
):
    """
    :title: Check the authentication of user after kdestroy and when ipa service stop.
    :setup:
        1. Add a user with --user-auth-type=passkey in the server with passkey mapping.
        2. Setup SSSD client with FIDO and umockdev, start SSSD service.
    :steps:
        1. Check authentication of the user and TGT after authentication.
        2. Remove the tgt using #kdestroy -A and stop the IPA service.
        3. Check the authentication again.
        4. Check that the user has been informed that the TGT ticket has not been granted.
    :expectedresults:
        1. User authenticates successfully and gets the TGT.
        2. Successfully remove the TGT and IPA is not reachable.
        3. User authenticate successfully, did not get TGT of user.
        4.  User has been correctly informed.
    :customerscenario: False
    """
    with open(f"{testdatadir}/passkey-mapping.ipa") as f:
        ipa.user("user1").add(user_auth_type="passkey").passkey_add(f.read().strip())

    client.sssd.start(service_user="root")

    rc, _, output, _ = client.auth.su.passkey_with_output(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.ipa",
        command="kdestroy -A",
    )

    assert rc == 0, "Authentication failed"
    ipa.svc.stop("ipa")

    rc, _, output, _ = client.auth.su.passkey_with_output(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.ipa",
        command="klist",
    )

    assert rc == 0, "Authentication failed"
    assert (
        "No Kerberos TGT granted as the server does not support this method. "
        "Your single-sign on(SSO) experience will be affected"
    ) in output, "Failed to get console message"
    klist_check = re.search(r"klist: Credentials cache.* not found", output)
    assert klist_check, "Credential cache found"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopology.IPA)
@pytest.mark.builtwith(client="passkey", ipa="passkey")
@pytest.mark.ticket(gh=6931)
def test_passkey__su_with_12_mappings(
    client: Client, ipa: IPA, moduledatadir: str, testdatadir: str, umockdev_ipaotpd_update
):
    """
    :title: Check authentication of user with IPA server when passkey mappings are 12 for a user
    :setup:
        1. Add a user with --user-auth-type=passkey in the server with 12 passkey mappings.
        2. Setup SSSD client with FIDO and umockdev, start SSSD service.
    :steps:
        1. Check authentication of the user.
        2. Check the TGT of user.
        3. Check that the user isn't informed about the degraded user experience due to not obtaining the TGT ticket.
    :expectedresults:
        1. User authenticates successfully.
        2. Get TGT after authentication of user.
        3. Not getting the message after authentication.
    :customerscenario: False
    """
    user_add = ipa.user("user1").add(user_auth_type="passkey")

    for n in range(1, 13):
        with open(f"{testdatadir}/passkey-mapping.ipa{n}") as f:
            user_add.passkey_add(f.read().strip())

    client.sssd.start(service_user="root")

    rc, _, output, _ = client.auth.su.passkey_with_output(
        username="user1",
        pin=123456,
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.ipa",
        command="klist",
    )

    assert rc == 0, "Authentication failed"
    assert "Ticket cache" in output, "Failed to get the TGT"
    assert (
        not (
            "No Kerberos TGT granted as the server does not support this method. "
            "Your single-sign on(SSO) experience will be affected"
        )
        in output
    ), "Get the console message about TGT"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopology.IPA)
@pytest.mark.builtwith(client="passkey", ipa="passkey")
@pytest.mark.ticket(gh=6931)
def test_passkey__su_no_pin_set(
    client: Client, ipa: IPA, moduledatadir: str, testdatadir: str, umockdev_ipaotpd_update
):
    """
    :title: Check authentication of user with IPA server when no pin set for the Passkey
    :setup:
        1. Add a user with --user-auth-type=passkey in the IPA server
        2. Modify Passkey configuration to set require user verification during authentication to false
        3. Setup SSSD client with FIDO and umockdev, start SSSD service
    :steps:
         1. Check authentication of the user when no pin set for the Passkey
         2. Check the TGT of user
    :expectedresults:
        1. User authenticates successfully
        2. Get TGT after authentication of user
    :customerscenario: False
    """
    with open(f"{testdatadir}/passkey-mapping.ipa") as f:
        ipa.user("user1").add(user_auth_type="passkey").passkey_add(f.read().strip())

    ipa.host.conn.run("ipa passkeyconfig-mod --require-user-verification=False", raise_on_error=False)
    client.sssd.start(service_user="root")

    rc, _, output, _ = client.auth.su.passkey_with_output(
        username="user1",
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.ipa",
        command="klist",
        auth_method=PasskeyAuthenticationUseCases.PASSKEY_NO_PIN_NO_PROMPTS,
    )

    assert rc == 0, "Authentication failed"
    assert "Ticket cache" in output, "Failed to get the TGT"
    assert not (
        (
            "No Kerberos TGT granted as the server does not support this method. "
            "Your single-sign on(SSO) experience will be affected"
        )
        in output
    ), "Got the console message about No Kerberos TGT granted"


@pytest.mark.importance("medium")
@pytest.mark.topology(KnownTopology.IPA)
@pytest.mark.builtwith(client="passkey", ipa="passkey")
@pytest.mark.ticket(gh=6931)
def test_passkey__prompt_options(
    client: Client, ipa: IPA, moduledatadir: str, testdatadir: str, umockdev_ipaotpd_update
):
    """
    :title: Check authentication of user with updated prompting options
    :setup:
        1. Add a user in the server with passkey mappings
        2. Add the prompting options to sssd.conf file
        3. Setup SSSD client with FIDO and umockdev, start SSSD service
    :steps:
        1. Check authentication of the user
        2. Check the updated prompt options
    :expectedresults:
        1. User authenticates successfully
        2. Got the updated prompt options
    :customerscenario: False
    """
    with open(f"{testdatadir}/passkey-mapping.ipa") as f:
        ipa.user("user1").add(user_auth_type="passkey").passkey_add(f.read().strip())

    client.sssd.section("prompting/passkey")["interactive"] = "True"
    client.sssd.section("prompting/passkey")["interactive_prompt"] = "Please, insert the passkey and press enter"
    client.sssd.section("prompting/passkey")["touch"] = "True"
    client.sssd.section("prompting/passkey")["touch_prompt"] = "Can you touch the passkey"
    client.sssd.start(service_user="root")

    rc, _, output, _ = client.auth.su.passkey_with_output(
        username="user1",
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.ipa",
        pin=123456,
        interactive_prompt="Please, insert the passkey and press enter",
        touch_prompt="Can you touch the passkey",
        command="klist",
        auth_method=PasskeyAuthenticationUseCases.PASSKEY_PIN_AND_PROMPTS,
    )

    assert rc == 0, "Authentication failed"
    assert "Ticket cache" in output, "Failed to get the TGT"
    assert (
        not (
            "No Kerberos TGT granted as the server does not support this method."
            "Your single-sign on(SSO) experience will be affected"
        )
        in output
    ), "Got the console message about No Kerberos TGT granted"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopology.IPA)
@pytest.mark.builtwith(client="passkey", ipa="passkey")
@pytest.mark.ticket(gh=7143)
def test_passkey__su_fallback_to_password(
    client: Client, ipa: IPA, moduledatadir: str, testdatadir: str, umockdev_ipaotpd_update
):
    """
    :title: Check password authentication of user with IPA server when sssd fall back to password authentication
    :setup:
        1. Add a user with --user-auth-type=passkey, password in the IPA server
        2. Setup SSSD client with FIDO and umockdev, start SSSD service
    :steps:
        1. Check authentication of the user with password
        2. Check the TGT of user
    :expectedresults:
        1. User authenticates successfully
        2. Get TGT after authentication of user
    :customerscenario: False
    """
    with open(f"{testdatadir}/passkey-mapping.ipa") as f:
        ipa.user("user1").add(user_auth_type=["passkey", "password"]).passkey_add(f.read().strip())

    client.sssd.start(service_user="root")

    rc, _, output, _ = client.auth.su.passkey_with_output(
        username="user1",
        device=f"{moduledatadir}/umockdev.device",
        ioctl=f"{moduledatadir}/umockdev.ioctl",
        script=f"{testdatadir}/umockdev.script.ipa",
        pin="\\n",
        command="klist",
        auth_method=PasskeyAuthenticationUseCases.PASSKEY_FALLBACK_TO_PASSWORD,
    )

    assert rc == 0, "Authentication failed"
    assert "Ticket cache" in output, "Failed to get the TGT"
    assert (
        not (
            "No Kerberos TGT granted as the server does not support this method."
            " Your single-sign on(SSO) experience will be affected"
        )
        in output
    ), "Got the console message about No Kerberos TGT granted"



"""
Infopipe tests.

:requirement: Service/IFP
"""

from __future__ import annotations

import pytest
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.generic import GenericProvider
from sssd_test_framework.topology import KnownTopology, KnownTopologyGroup


@pytest.mark.topology(KnownTopology.LDAP)
def test_infopipe__get_user_properties(client: Client, provider: GenericProvider):
    """
    :title: Get a user's property
    :setup:
        1. Add 'user-1' to the provider
        2. Launch SSSD
    :steps:
        1. Call FindByName() to get the user's object path
        2. Create a new object associated to that user using its object path
        3. Get the UID property using Get()
        4. Get the GID property with direct access
        5. Get all the properties
    :expectedresults:
        1. User is found and object path retrieved
        2. The object is created
        3. Property is returned and has the right value
        4. Property is returned and has the right value
        5. Some sample properties are present and have the right value
    :customerscenario: False
    """
    provider.user("user-1").add(uid=10001, gid=20001)

    client.sssd.start()

    users = client.ifp.getObject("/org/freedesktop/sssd/infopipe/Users")
    user_path = users.FindByName("user-1")
    assert (
        "/org/freedesktop/sssd/infopipe/Users/test/10001" == user_path
    ), "FindByName('user-1') returned an unexpected user path"

    user = client.ifp.getObject(user_path)
    uid = user.Get("org.freedesktop.sssd.infopipe.Users.User", "uidNumber")
    assert uid == 10001, "user.Get(user_path, 'uidNumber') returned an unexpected UID"

    gid = user.gidNumber
    assert gid == 20001, "user.gidNumber returned an unexpected GID"

    props = user.GetAll("org.freedesktop.sssd.infopipe.Users.User")
    assert props["uidNumber"] == 10001, "GetAll(Users.User) returned an unexpected uidNumber"
    assert props["gidNumber"] == 20001, "GetAll(Users.User) returned an unexpected gidNumber"
    assert props["name"] == "user-1", "GetAll(Users.User) returned an unexpected name"
    assert props["homeDirectory"] == "/home/user-1", "GetAll(Users.User) returned an unexpected homeDirectory"


@pytest.mark.topology(KnownTopology.LDAP)
def test_infopipe__get_domain_properties(client: Client):
    """
    :title: Access a domain's information through InfoPipe
    :setup:
        1. Launch SSSD
    :steps:
        1. Call IsOnline()
        2. Call GetAll() with "org.freedesktop.sssd.infopipe.Domains"
        3. Read the property enumerable
    :expectedresults:
        1. Domain is online
        2. All data is returned
        3. Is enumerable
    :customerscenario: False
    """
    client.sssd.start()

    domain = client.ifp.getObject("/org/freedesktop/sssd/infopipe/Domains/test")

    res = domain.IsOnline()
    assert isinstance(res, bool), "IsOnline() didn't return a boolean"
    assert res, "IsOnline() returned false"

    res = domain.GetAll("org.freedesktop.sssd.infopipe.Domains")
    assert isinstance(res, dict), "GetAll(infopipe.Domains) didn't return a dict"
    assert res["name"] == "test", "GetAll(infopipe.Domains) returned an unexpected name"
    assert res["provider"] == "ldap", "GetAll(infopipe.Domains) returned an unexpected provider"
    assert (
        res["login_format"] == r"^((?P<name>.+)@(?P<domain>[^@]+)|(?P<name>[^@]+))$"
    ), "GetAll(infopipe.Domains) returned an unexpected login_format"
    assert isinstance(
        res["primary_servers"], list
    ), "GetAll(infopipe.Domains) returned a primary_servers which is not a list"
    assert (
        len(res["primary_servers"]) == 1
    ), "GetAll(infopipe.Domains) returned a wrong number of elements in primary_servers"
    assert (
        res["primary_servers"][0] == "ldap://master.ldap.test"
    ), "GetAll(infopipe.Domains) returned a wrong URI in primary_servers"

    res = domain.enumerable
    assert isinstance(res, bool), "Property enumerable didn't return a boolean"
    assert not res, "Property enumerable returned the wrong value"


@pytest.mark.ticket(gh=6020, bz=2128840, jira="SSSD-5054")
@pytest.mark.topology(KnownTopology.IPA)
@pytest.mark.topology(KnownTopology.LDAP)
def test_infopipe__list_by_attr(client: Client, provider: GenericProvider):
    """
    :title: Test InfoPipe.Users.ListByAttr()
    :setup:
        1. Add 'user-1' to the provider
        2. Add 'user-2' to the provider
        3. Add 'user-3' to the provider
        4. Add "user_attributes = +extraName" to the [ifp] configuration section
        5. Add "ldap_user_extra_attrs = extraName:uid" to the 'test' domain configuration
        6. Create the 'app' application domain inheriting from 'test'
        7. Launch SSSD
    :steps:
        1. Call ListByAttr() with "extraName=user-2" and no limit
        2. Call ListByAttr() with "extraName=user-*" and no limit
        3. Call ListByAttr() with "extraName=user-*" and a limit of 4
        4. Call ListByAttr() with "extraName=nouser-*" and no limit
        5. Call ListByAttr() with "noName=*" and no limit
    :expectedresults:
        1. User is found exactly once for 'test' and once for 'app'
        2. All three users are found on each domain
        3. Four of the six user entries are returned
        4. No user is returned
        5. No user is returned
    :customerscenario: False
    """
    provider.user("user-1").add(uid=10001)
    provider.user("user-2").add(uid=10002)
    provider.user("user-3").add(uid=10003)

    cfg = client.sssd.config
    cfg.set("sssd", "domains", cfg.get("sssd", "domains") + ",app")
    cfg.add_section("ifp")
    cfg.set("ifp", "user_attributes", "+extraName")
    cfg.set("domain/test", "ldap_user_extra_attrs", "extraName:uid")
    cfg.add_section("application/app")
    cfg.set("application/app", "inherit_from", "test")

    client.sssd.start()

    users = client.ifp.getObject("/org/freedesktop/sssd/infopipe/Users")

    list = users.ListByAttr("extraName", "user-2", 0)
    assert len(list) == 2, "ListByAttr('extraName', 'user-2', 0) returned a wrong number of elements"
    assert (
        "/org/freedesktop/sssd/infopipe/Users/test/10002" in list
    ), "ListByAttr('extraName', 'user-2', 0) is missing element test/user-2"
    assert (
        "/org/freedesktop/sssd/infopipe/Users/app/user_2d2_40app" in list
    ), "ListByAttr('extraName', 'user-2', 0) is missing element app/user-2"

    list = users.ListByAttr("extraName", "user-*", 0)
    assert len(list) == 6, "Wrong number of elements in list"
    assert (
        "/org/freedesktop/sssd/infopipe/Users/test/10001" in list
    ), "ListByAttr('extraName', 'user-*', 0) is missing element test/10001"
    assert (
        "/org/freedesktop/sssd/infopipe/Users/test/10002" in list
    ), "ListByAttr('extraName', 'user-*', 0) is missing element test/10002"
    assert (
        "/org/freedesktop/sssd/infopipe/Users/test/10003" in list
    ), "ListByAttr('extraName', 'user-*', 0) is missing element test/10003"
    assert (
        "/org/freedesktop/sssd/infopipe/Users/app/user_2d1_40app" in list
    ), "ListByAttr('extraName', 'user-*', 0) is missing element app/user-1"
    assert (
        "/org/freedesktop/sssd/infopipe/Users/app/user_2d2_40app" in list
    ), "ListByAttr('extraName', 'user-*', 0) is missing element app/user-2"
    assert (
        "/org/freedesktop/sssd/infopipe/Users/app/user_2d3_40app" in list
    ), "ListByAttr('extraName', 'user-*', 0) is missing element app/user-3"

    model = list  # We'll use this list as model for the next check
    list = users.ListByAttr("extraName", "user-*", 4)
    assert len(list) == 4, "ListByAttr('extraName', 'user-*', 4) returned a wrong number of elements"
    # We don't know which two will be returned
    for user in list:
        assert user in model, "ListByAttr('extraName', 'user-*', 4) returned an unexpected element"

    list = users.ListByAttr("extraName", "nouser-*", 0)
    assert len(list) == 0, "ListByAttr('extraName', 'nouser-*', 0) returned unexpected elements"

    list = users.ListByAttr("noattr", "*", 0)
    assert len(list) == 0, "ListByAttr('noattr', '*', 0) returned unexpected elements"


@pytest.mark.ticket(gh=[6360, 6361], jira="SSSD-5054")
@pytest.mark.topology(KnownTopology.IPA)
@pytest.mark.topology(KnownTopology.LDAP)
def test_infopipe__list_by_name(client: Client, provider: GenericProvider):
    """
    :title: Test InfoPipe.Users.ListByName()
    :setup:
        1. Add 'user-1' to the provider
        2. Add 'user-2' to the provider
        3. Add 'user-3' to the provider
        4. Launch SSSD
    :steps:
        1. Call ListByName() with "user-2" and no limit
        2. Call ListByName() with "user-*" and no limit
        3. Call ListByName() with "user-*" and a limit of 2
        4. Call ListByName() with "nouser-*" and no limit
    :expectedresults:
        1. User is found and none else
        2. All three users are found
        3. Two of the three users are returned
        4. No user is returned
    :customerscenario: False
    """
    provider.user("user-1").add(uid=10001)
    provider.user("user-2").add(uid=10002)
    provider.user("user-3").add(uid=10003)
    client.sssd.start()

    users = client.ifp.getObject("/org/freedesktop/sssd/infopipe/Users")

    result = users.ListByName("user-2", 0)
    assert (
        "/org/freedesktop/sssd/infopipe/Users/test/10002" in result
    ), "ListByName('user-2', 0) returned unexpected element"
    assert len(result) == 1, "ListByName('user-2', 0) returned a wrong number of elements"

    result = users.ListByName("user-*", 0)
    assert (
        "/org/freedesktop/sssd/infopipe/Users/test/10001" in result
    ), "ListByName('user-*', 0) is missing element 10001"
    assert (
        "/org/freedesktop/sssd/infopipe/Users/test/10002" in result
    ), "ListByName('user-*', 0) is missing element 10002"
    assert (
        "/org/freedesktop/sssd/infopipe/Users/test/10003" in result
    ), "ListByName('user-*', 0) is missing element 10003"
    assert len(result) == 3, "ListByName('user-*', 0) returned a wrong number of elements"

    model = result  # We'll use this list as model for the next check
    result = users.ListByName("user-*", 2)
    assert len(result) == 2, "ListByName('user-*', 2) returned a wrong number of elements"
    # We don't know which two will be returned
    for user in result:
        assert user in model, "ListByName('user-*', 2) returned an unexpected element"

    result = users.ListByName("nouser*", 0)
    assert len(result) == 0, "ListByName('nouser*', 0) returned unexpected elements"


@pytest.mark.importance("medium")
@pytest.mark.ticket(bz=1667252)
@pytest.mark.topology(KnownTopologyGroup.AnyProvider)
def test_infopipe__lookup_user_with_extra_attributes(client: Client, provider: GenericProvider):
    """
    :title: Infopipe does not crash looking up extra attribute
    :setup:
        1. Create user "user1"
        2. Enable infopipe, add a test attribute and start SSSD
    :steps:
        1. Lookup user using sssctl
        2. Check SSSD service
    :expectedresults:
        1. User found
        2. Service is running
    :customerscenario: True
    """
    provider.user("user1").add()
    client.sssd.sssd["services"] = "nss, pam, ifp"
    client.sssd.domain["ldap_user_extra_attrs"] = "test:homeDirectory"
    client.sssd.ifp["user_attributes"] = "+test"
    client.sssd.start()

    result = client.sssctl.user_checks("user1")
    assert result.rc == 0, "User not found!"

    result = client.sssd.svc.status("sssd")
    assert result.rc == 0, "Service is not running!"
